from src.download.modules import *
from src.renum.shared.handling_chain_numbering_clashes import handling_chain_numbering_clashes
from src.renum.shared.SIFTS_tree_parser import SIFTS_tree_parser
from src.renum.shared.renumbered_count_in_chains import renumbered_count_in_chains
from src.download.downloadwithThreadPool import download_with_pool, url_formation_for_pool
PDBrenum_REMARK = [
    "REMARK   0  File processed by PDBrenum: http://dunbrack3.fccc.edu/PDBrenum      ",
    "REMARK   0  Author sequence numbering is replaced with UniProt numbering        ",
    "REMARK   0  according to alignment by SIFTS                                     ",
    "REMARK   0  (https://www.ebi.ac.uk/pdbe/docs/sifts/).                           ",
    "REMARK   0  Only chains with UniProt sequences in SIFTS are renumbered.         ",
    "REMARK   0  Residues in UniProt chains without UniProt residue numbers in SIFTS ",
    "REMARK   0  (e.g., sequence tags) are given residue numbers 5000+label_seq_id   ",
    "REMARK   0  (where label_seq_id is the 1-to-N residue numbering of each chain.  ",
    "REMARK   0  Ligands are numbered 5000+their residue number in the original      ",
    "REMARK   0  file. The _poly_seq_scheme table contains a correspondence between  ",
    "REMARK   0  the 1-to-N sequence (seq_id), the new numbering based on UniProt    ",
    "REMARK   0  (pdb_seq_num = auth_seq_id in the _atom_site records), and          ",
    "REMARK   0  the author numbering in the original mmCIF file from the PDB        ",
    "REMARK   0  (auth_seq_num).                                                     "]


def SIFTS_data_parser_for_PDB(tuple_PDBe_for_PDB_and_tuple_PDB, tuple_PDBe_for_UniProt_and_tuple_UniProt,
                              default_PDB_num, chains_to_change="all"):
    df_PDBe_UniProt = pd.DataFrame(tuple_PDBe_for_UniProt_and_tuple_UniProt, columns=['PDBe', 'UniProt', "AccessionID"])
    df_PDBe_UniProt = df_PDBe_UniProt.drop_duplicates(subset="PDBe", keep='first')
    df_PDBe_PDB = pd.DataFrame(tuple_PDBe_for_PDB_and_tuple_PDB, columns=['PDBe', 'PDB'])
    df_PDBe_PDB = df_PDBe_PDB.drop_duplicates(subset="PDBe", keep='first')

    df_PDBe_PDB_UniProt = df_PDBe_PDB.merge(df_PDBe_UniProt, left_on="PDBe", right_on="PDBe", how='left')
    df_PDBe_PDB_UniProt['UniProt'] = df_PDBe_PDB_UniProt['UniProt'].replace(np.nan, "5000")
    df_PDBe_PDB_UniProt["Uni_moD"] = np.where(df_PDBe_PDB_UniProt['UniProt'] != "5000", df_PDBe_PDB_UniProt['UniProt'], df_PDBe_PDB_UniProt["PDBe"])
    df_PDBe_PDB_UniProt.loc[:, 'new_col_Uni'] = df_PDBe_PDB_UniProt.Uni_moD.map(lambda x: x[0])
    df_PDBe_PDB_UniProt["UniProt_5k"] = df_PDBe_PDB_UniProt.new_col_Uni.apply(lambda x: (int(x) + default_PDB_num if type(x) == str else x))
    df_PDBe_PDB_UniProt.loc[df_PDBe_PDB_UniProt['UniProt'] != '5000', 'UniProt_5k'] = df_PDBe_PDB_UniProt['new_col_Uni']

    Three_Rows_CIF_Num_Uni = []
    if chains_to_change == "all":
        for index, rows in df_PDBe_PDB_UniProt.iterrows():
            intermediate_list = [rows.PDBe, rows.UniProt_5k, rows.Uni_moD, rows.PDB, rows.AccessionID]
            Three_Rows_CIF_Num_Uni.append(intermediate_list)

    else:
        for index, rows in df_PDBe_PDB_UniProt.iterrows():
            if rows.PDB[2].strip() in chains_to_change:
                intermediate_list = [rows.PDBe, rows.UniProt_5k, rows.Uni_moD, rows.PDB, rows.AccessionID]
            else:
                intermediate_list = [rows.PDBe, rows.PDB[0], rows.Uni_moD, rows.PDB, rows.AccessionID]
            Three_Rows_CIF_Num_Uni.append(intermediate_list)

    df_PDBe_PDB_UniProt["Three_Rows_CIF_Num_Uni"] = Three_Rows_CIF_Num_Uni
    df_PDBe_PDB_UniProt_without_null = df_PDBe_PDB_UniProt[df_PDBe_PDB_UniProt.PDB.map(lambda x: x[0]) != "null"]
    df_PDBe_PDB_UniProt_without_null_index_PDBe = df_PDBe_PDB_UniProt_without_null.set_index("PDBe")

    return [df_PDBe_PDB_UniProt_without_null_index_PDBe, df_PDBe_PDB_UniProt]


def try_SIFTS_tree_parser(default_input_path_to_SIFTS, SIFTS_name):
    product_tree_SIFTS = 0
    for _ in range(3):
        try:
            product_tree_SIFTS = SIFTS_tree_parser(
                gzip.open(Path(str(default_input_path_to_SIFTS) + "/" + SIFTS_name), 'rt'))
            break
        except EOFError:
            os.remove(Path(str(default_input_path_to_SIFTS) + "/" + SIFTS_name))
            download_with_pool(url_formation_for_pool("SIFTS", [SIFTS_name], default_input_path_to_SIFTS=default_input_path_to_SIFTS)[0],
                               default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        except ValueError:
            os.remove(Path(str(default_input_path_to_SIFTS) + "/" + SIFTS_name))
            download_with_pool(url_formation_for_pool("SIFTS", [SIFTS_name], default_input_path_to_SIFTS=default_input_path_to_SIFTS)[0],
                               default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        except OSError:
            download_with_pool(url_formation_for_pool("SIFTS", [SIFTS_name], default_input_path_to_SIFTS=default_input_path_to_SIFTS)[0],
                               default_input_path_to_SIFTS=default_input_path_to_SIFTS)
    return product_tree_SIFTS


def try_PDB(default_input_path_to_PDB, PDB):
    split = 0

    for _ in range(3):
        try:
            split = gzip.open(Path(str(default_input_path_to_PDB) + "/" + PDB), 'rt').read().splitlines()
            break
        except EOFError:
            try:
                re.search('\.pdb(.*).gz', PDB).group(1)
                os.remove(Path(str(default_input_path_to_PDB) + "/" + PDB))
                download_with_pool(url_formation_for_pool("PDB_assembly", [PDB], default_input_path_to_PDB_assembly=default_input_path_to_PDB)[0],
                                   default_input_path_to_PDB_assembly=default_input_path_to_PDB)
            except AttributeError:
                os.remove(Path(str(default_input_path_to_PDB) + "/" + PDB))
                download_with_pool(url_formation_for_pool("PDB", [PDB], default_input_path_to_PDB=default_input_path_to_PDB)[0],
                                   default_input_path_to_PDB=default_input_path_to_PDB)

        except ValueError:
            try:
                re.search('\.pdb(.*).gz', PDB).group(1)
                os.remove(Path(str(default_input_path_to_PDB) + "/" + PDB))
                download_with_pool(url_formation_for_pool("PDB_assembly", [PDB], default_input_path_to_PDB_assembly=default_input_path_to_PDB)[0],
                                   default_input_path_to_PDB_assembly=default_input_path_to_PDB)
            except AttributeError:
                os.remove(Path(str(default_input_path_to_PDB) + "/" + PDB))
                download_with_pool(url_formation_for_pool("PDB", [PDB], default_input_path_to_PDB=default_input_path_to_PDB)[0],
                                   default_input_path_to_PDB=default_input_path_to_PDB)
        except OSError:
            try:
                re.search('\.pdb(.*).gz', PDB).group(1)
                download_with_pool(url_formation_for_pool("PDB_assembly", [PDB], default_input_path_to_PDB_assembly=default_input_path_to_PDB)[0],
                                   default_input_path_to_PDB_assembly=default_input_path_to_PDB)
            except AttributeError:
                download_with_pool(url_formation_for_pool("PDB", [PDB], default_input_path_to_PDB=default_input_path_to_PDB)[0],
                                   default_input_path_to_PDB=default_input_path_to_PDB)
    return split


def if_no_SIFTS_data_log_for_PDB(default_input_path_to_PDB, PDB_id, PDB):
    split = try_PDB(default_input_path_to_PDB, PDB)
    res_number_name_chainID_from_PDB_tuple = list()
    chains_set = set()
    log_message = list()

    for n in split:
        if n.startswith("ATOM") or n.startswith("TER") or n.startswith("ANISOU") or n.startswith("ANISOU") or n.startswith("SIGUIJ"):
            res_number_name_chainID_from_PDB_tuple.append((n[22:27].strip(" "), n[17:20], n[21]))
            chains_set.add(n[21])

    if len(res_number_name_chainID_from_PDB_tuple) == 0:
        log_message.append([PDB_id, "-", "-", "-", "-", "-", "-", "-", "-", "-"])
        return log_message

    df_mmCIF = pd.DataFrame(list(zip(res_number_name_chainID_from_PDB_tuple, res_number_name_chainID_from_PDB_tuple)))
    df_mmCIF = df_mmCIF.rename(columns={0: "PDB_old", 1: "PDB_old_copy"})
    df_mmCIF = df_mmCIF.set_index("PDB_old")
    df_mmCIF = df_mmCIF.drop_duplicates()

    for chain in chains_set:
        count_res_in_chain = 0
        for resnum_resname_chain in df_mmCIF.PDB_old_copy:
            if chain == resnum_resname_chain[2]:
                count_res_in_chain += 1
        log_message.append([PDB_id, "-", chain, "-", "-", len(df_mmCIF), "-", count_res_in_chain, "0", "0"])
    return log_message


def copy_file(inpath, file_name, outpath, postfix, gzip_mode):
    if file_name.endswith(".ent.gz") and file_name.startswith("pdb"):
        PDB_id = file_name[3:file_name.rfind(".ent.gz")]
    else:
        PDB_id = file_name[:4]

    absolute_path_in = inpath + "/" + file_name
    absolute_path_out = outpath + "/" + PDB_id + postfix
    if gzip_mode == "off":
        with gzip.open(absolute_path_in, 'rb') as f_in:
            with open(absolute_path_out, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
    else:
        shutil.copyfile(absolute_path_in, absolute_path_out + ".gz")


def PDB_parser(split, df_PDBe_PDB_UniProt_without_null_index_PDBe, default_PDB_num):
    res_number_name_chainID_from_PDB_tuple = list()
    num_ins_code_name_chain_HETATM = list()
    missing_res_remark_465 = list()
    skipper_for_remark_465 = True
    Num_is_Too_Big = True

    for n in split:
        if n.startswith("ATOM") or n.startswith("TER") or n.startswith("ANISOU") or n.startswith("ANISOU") or n.startswith("SIGUIJ"):
            res_number_name_chainID_from_PDB_tuple.append((n[22:27].strip(" "), n[17:20], n[21]))
        if n.startswith("HETATM"):
            num_ins_code_name_chain_HETATM.append((n[22:27].strip(" "), n[17:20], n[21]))

        if n.startswith("REMARK 465"):
            if not skipper_for_remark_465:
                missing_res_remark_465.append((n[20:27].strip(" "), n[15:18], n[19]))
            if n[15:27] == "RES C SSSEQI":
                skipper_for_remark_465 = False

    df_mmCIF = pd.DataFrame(list(zip(res_number_name_chainID_from_PDB_tuple, res_number_name_chainID_from_PDB_tuple)))
    df_mmCIF = df_mmCIF.rename(columns={0: "PDB_old", 1: "PDB_old_copy"})
    df_mmCIF = df_mmCIF.set_index("PDB_old")
    df_mmCIF = df_mmCIF.drop_duplicates()

    df_final = df_mmCIF.merge(df_PDBe_PDB_UniProt_without_null_index_PDBe, left_on="PDB_old_copy", right_on="PDB", how='left')
    df_final['Uni_moD'] = df_final['Uni_moD'].replace(np.nan, "5000")
    df_final["Uni_moD"] = np.where(df_final['Uni_moD'] != "5000", df_final['Uni_moD'], df_final["PDB_old_copy"])
    df_final.loc[:, 'new_col_Uni'] = df_final.Uni_moD.map(lambda x: x[0])
    df_final["UniProt_5k"] = df_final.new_col_Uni.apply(lambda x: (int(x) + default_PDB_num if x.isdigit() else x))
    df_final.loc[df_final['UniProt'] != '5000', 'UniProt_5k'] = df_final['new_col_Uni']

    Three_Rows_CIF_Num_Uni = list()
    for index, rows in df_final.iterrows():
        intermediate_list = [rows.PDB_old_copy, rows.UniProt_5k, rows.Uni_moD]
        if type(rows.UniProt_5k) == int:
            if len(str(rows.UniProt_5k)) > 4:
                Num_is_Too_Big = False
        else:
            if len(rows.UniProt_5k) > 4:
                Num_is_Too_Big = False
        Three_Rows_CIF_Num_Uni.append(intermediate_list)

    df_final["Three_Rows_CIF_Num_Uni"] = Three_Rows_CIF_Num_Uni
    df_final_index_PDBe = df_final.set_index("PDB_old_copy")
    df_final_index_PDBe_drop_NAN = df_final_index_PDBe.dropna(subset=['PDB'])
    pd_series_index_PDBe = df_final_index_PDBe["Three_Rows_CIF_Num_Uni"]

    PDB_str = df_final_index_PDBe_drop_NAN.PDB.map(
        lambda x: x[1] + "{:>2}".format(x[2]) + "{:>4}".format(x[0]) + " " if x[0].isdigit() else x[1] + "{:>2}".format(x[2]) + "{:>5}".format(x[0]))
    df_final_index_PDBe_drop_NAN = df_final_index_PDBe_drop_NAN.merge(PDB_str.rename('PDB_str'), left_index=True, right_index=True)
    df_final_poly_corrected = df_final_index_PDBe_drop_NAN.drop(columns=['UniProt', 'AccessionID', "new_col_Uni", "UniProt_5k", "Uni_moD"])
    renum_str = df_final_poly_corrected.PDB.map(lambda x: x[1] + "{:>2}".format(x[2])) + df_final_poly_corrected.Three_Rows_CIF_Num_Uni.map(
        lambda x: "{:>4}".format(str(int(x[1])))) + " "
    df_final_poly_corrected = df_final_poly_corrected.merge(renum_str.rename('renum_str'), left_index=True, right_index=True)
    df_final_poly_corrected = df_final_poly_corrected.reset_index(drop=True)

    return [pd_series_index_PDBe, num_ins_code_name_chain_HETATM, df_final_poly_corrected, missing_res_remark_465, Num_is_Too_Big]


def non_poly_num(pd_series_index_PDBe, num_ins_code_name_chain_HETATM):
    working_range_list = list()
    for n in range(1, 10000):
        working_range_list.append(n)

    chain_and_number = list()
    for n in pd_series_index_PDBe:
        chain_and_number.append((n[0][2], n[1]))

    chain_label = chain_and_number[0][0]
    numbers_per_chain = list()
    chain_label_with_numbers_per_chain = list()
    for n in chain_and_number:
        if chain_label == n[0]:
            d = str(n[1])
            n_numeric = ''.join(d for d in d if d.isdigit())
            numbers_per_chain.append(int(n_numeric))
        else:
            numbers_per_chain = list(np.unique(numbers_per_chain))
            chain_label_with_numbers_per_chain.append((chain_label, numbers_per_chain))
            numbers_per_chain = list()
            chain_label = n[0]
            if chain_label == n[0]:
                d = str(n[1])
                n_numeric = ''.join(d for d in d if d.isdigit())
                numbers_per_chain.append(int(n_numeric))

    numbers_per_chain = list(np.unique(numbers_per_chain))
    chain_label_with_numbers_per_chain.append((chain_label, numbers_per_chain))

    available_numbers_for_chains = list()
    for n in chain_label_with_numbers_per_chain:
        for num in n[1]:
            if num in working_range_list:
                working_range_list.remove(num)

        available_numbers_for_chains.append((n[0], working_range_list))
        working_range_list = list()
        for n_ in range(1, 10000):
            working_range_list.append(n_)

    chain_and_num_available = list()
    for n in available_numbers_for_chains:
        for num in n[1]:
            chain_and_num_available.append((n[0], num))

    numbers_from_num_ins_code_name_chain_HETATM = list()
    chain_from_num_ins_code_name_chain_HETATM = list()
    for n in num_ins_code_name_chain_HETATM:
        numbers_from_num_ins_code_name_chain_HETATM.append(n[0])
        chain_from_num_ins_code_name_chain_HETATM.append(n[2])

    df_nonpoly = pd.DataFrame(
        list(zip(num_ins_code_name_chain_HETATM, chain_from_num_ins_code_name_chain_HETATM, numbers_from_num_ins_code_name_chain_HETATM)), columns=[
            'PDB', "PDB_chain", "numbers"])
    df_nonpoly_dropped_dup = df_nonpoly.drop_duplicates(subset="PDB", keep='first')
    df_nonpoly_dropped_dup_sorted = df_nonpoly_dropped_dup.sort_values(["PDB_chain", "numbers"], ascending=(True, True)).reset_index(drop=True)
    small_ref_table = df_nonpoly_dropped_dup_sorted.set_index(["PDB_chain", "PDB"]).count(level="PDB_chain")

    all_nonpoly_chains = list()
    for n in small_ref_table.index:
        all_nonpoly_chains.append(n)

    checked_chains_list = list()
    for n in chain_and_num_available:
        checked_chains_list.append(n[0])
    checked_chains_list_uniq = list(np.unique(checked_chains_list))

    available_numbers_to_chains = list()
    for n in all_nonpoly_chains:
        if n not in checked_chains_list_uniq:
            for z in range(1, 10000):
                available_numbers_to_chains.append((n, z))

    chain_and_num_available.extend(available_numbers_to_chains)

    df_chain_and_num_available = pd.DataFrame(chain_and_num_available, columns=['available_chain', "available_number"])
    df_chain_and_num_available_sorted = df_chain_and_num_available.drop_duplicates(
        subset=["available_chain", "available_number"], keep="first").sort_values(
        ["available_chain", "available_number"], ascending=(True, False)).reset_index(drop=True)

    df_for_nonpoly_replace = pd.DataFrame(list(), columns=['available_chain', "available_number"])
    for n in df_nonpoly_dropped_dup_sorted.set_index(["PDB_chain", "PDB"]).count(level="PDB_chain").index:
        temporal_df_for_addition_of_available_num = df_chain_and_num_available_sorted.where(
            df_chain_and_num_available_sorted['available_chain'] == n).dropna()[0:(small_ref_table["numbers"][n])]
        df_for_nonpoly_replace = df_for_nonpoly_replace.append(temporal_df_for_addition_of_available_num, ignore_index=True)

    df_final_nonpoly = pd.merge(left=df_nonpoly_dropped_dup_sorted, right=df_for_nonpoly_replace, left_index=True, right_index=True)

    df_final_nonpoly.loc[:, 'PDB_str'] = df_final_nonpoly.PDB.map(
        lambda x: x[1] + "{:>2}".format(x[2]) + "{:>4}".format(x[0]) + " " if x[0].isdigit() else x[1] + "{:>2}".format(x[2]) + "{:>5}".format(x[0]))
    df_final_nonpoly_corrected = df_final_nonpoly.drop(columns=['PDB_chain', 'numbers', "available_chain"])
    df_final_nonpoly_corrected.loc[:, 'renum_str'] = df_final_nonpoly_corrected.PDB.map(
        lambda x: x[1] + "{:>2}".format(x[2])) + df_final_nonpoly_corrected.available_number.map(lambda x: "{:>4}".format(str(int(x)))) + " "

    return df_final_nonpoly_corrected


def remark_465(missing_res_remark_465, df_PDBe_PDB_UniProt):
    df_PDBe_PDB_UniProt_nulls = df_PDBe_PDB_UniProt.loc[df_PDBe_PDB_UniProt['PDB'].apply(lambda x: x[0] == "null")]
    df_PDBe_PDB_UniProt_nulls = df_PDBe_PDB_UniProt_nulls.reset_index(drop=True)

    df_mmCIF_remark_465 = pd.DataFrame(list(zip(missing_res_remark_465)))
    df_mmCIF_remark_465 = df_mmCIF_remark_465.rename(columns={0: "PDB_old"})
    df_mmCIF_remark_465 = df_mmCIF_remark_465.drop_duplicates()

    df_remark_465_final = df_mmCIF_remark_465.merge(df_PDBe_PDB_UniProt_nulls, left_index=True, right_index=True)
    df_remark_465_final.loc[:, 'PDB_str'] = df_remark_465_final.PDB_old.map(
        lambda x: x[1] + "{:>2}".format(x[2]) + "{:>6}".format(x[0]) + " " if x[0].isdigit() else x[1] + "{:>2}".format(x[2]) + "{:>5}".format(x[0]))

    df_final_poly_remark_465_corrected = df_remark_465_final.drop(columns=['UniProt', 'AccessionID', "new_col_Uni", "UniProt_5k", "Uni_moD"])
    df_final_poly_remark_465_corrected.loc[:, 'renum_str'] = df_final_poly_remark_465_corrected.PDB_old.map(
        lambda x: x[1] + "{:>2}".format(x[2])) + df_final_poly_remark_465_corrected.Three_Rows_CIF_Num_Uni.map(
        lambda x: "{:>6}".format(str(int(x[1])))) + " "
    df_final_poly_remark_465_corrected = df_final_poly_remark_465_corrected.reset_index(drop=True)

    return df_final_poly_remark_465_corrected


def final_dict_formation(df_final_poly_corrected, df_final_nonpoly_corrected, final_remark_465, chains_to_change):
    all_data_df = df_final_poly_corrected.append(df_final_nonpoly_corrected, ignore_index=True, sort=False)
    all_data_df = all_data_df.append(final_remark_465, ignore_index=True, sort=False)
    all_data_df_drop_dup = all_data_df.drop_duplicates(subset="PDB_str", keep='first')
    not_in_chains_to_change = all_data_df_drop_dup.PDB.map(lambda x: x if x[2] in chains_to_change else "not_in_chains_to_change")
    all_data_merged_not_in_chain_to_change = all_data_df_drop_dup.merge(not_in_chains_to_change.rename('not_in_chains_to_change'),
                                                                        left_index=True, right_index=True)
    all_data_df_drop_dup_drop_chains = all_data_merged_not_in_chain_to_change[
        all_data_merged_not_in_chain_to_change.not_in_chains_to_change != 'not_in_chains_to_change']

    different_indent_PDB_str = list()
    for n in all_data_df_drop_dup_drop_chains["PDB_str"]:
        if "-" in n:
            n = n[:5] + n[6:] + " "
        different_indent_PDB_str.append(n)
        different_indent_PDB_str.append(n[:3] + " " + n[3:])  # present HET
        different_indent_PDB_str.append(n[:5] + " " + n[5:])  # present most common
        different_indent_PDB_str.append(n[:5] + "  " + n[5:])  # present REMARK 500

    different_indent_renum_str = list()
    for n in all_data_df_drop_dup_drop_chains["renum_str"]:
        different_indent_renum_str.append(n)
        different_indent_renum_str.append(n[:3] + " " + n[3:])  # present HET
        different_indent_renum_str.append(n[:5] + " " + n[5:])  # present most common
        different_indent_renum_str.append(n[:5] + "  " + n[5:])  # present REMARK 500

    dict_for_replacement = dict(zip(different_indent_PDB_str, different_indent_renum_str))
    return dict_for_replacement


def replace_all(lines, dict_for_replacement):
    location_of_the_value = 0
    for key, value in dict_for_replacement.items():
        if key in lines:
            if location_of_the_value == lines.find(key):
                continue
            lines = lines.replace(key, value)
            location_of_the_value = lines.find(value)
    return lines


def master_PDB_renumber_function(input_PDB_files_were_found, default_input_path_to_PDB, default_input_path_to_SIFTS,
                                 default_output_path_to_PDB, default_PDB_num, gzip_mode, exception_AccessionIDs):
    if not os.path.exists(default_output_path_to_PDB):
        os.makedirs(default_output_path_to_PDB)

    input_PDB_files_were_found_list = list()
    input_PDB_files_were_found_list.append(input_PDB_files_were_found)

    for PDB in input_PDB_files_were_found_list:

        try:
            assembly_num = re.search('\.pdb(.*).gz', PDB).group(1)
            SIFTS_name = PDB[:4] + ".xml.gz"
            PDB_id = PDB[:4]
        except AttributeError:
            assembly_num = ""
            SIFTS_name = PDB[3:7] + ".xml.gz"
            PDB_id = PDB[3:7]

        # for no corresponding SIFTS
        try:
            gzip.open(Path(str(default_input_path_to_SIFTS) + "/" + SIFTS_name), 'rt')
        except FileNotFoundError:
            copy_file(default_input_path_to_PDB, PDB, default_output_path_to_PDB, "_renum.pdb" + assembly_num, gzip_mode)
            log_message = if_no_SIFTS_data_log_for_PDB(default_input_path_to_PDB, PDB_id, PDB)
            return log_message

        # for zero byte SIFTS
        if os.path.getsize(Path(str(default_input_path_to_SIFTS) + "/" + SIFTS_name)) == 0:
            copy_file(default_input_path_to_PDB, PDB, default_output_path_to_PDB, "_renum.pdb" + assembly_num, gzip_mode)
            log_message = if_no_SIFTS_data_log_for_PDB(default_input_path_to_PDB, PDB_id, PDB)
            return log_message

        product_tree_SIFTS = try_SIFTS_tree_parser(default_input_path_to_SIFTS, SIFTS_name)
        if product_tree_SIFTS == 0:
            continue

        tuple_PDBe_for_PDB_and_tuple_PDB = product_tree_SIFTS[0]
        tuple_PDBe_for_UniProt_and_tuple_UniProt = product_tree_SIFTS[1]
        UniProt_conversion_dict = product_tree_SIFTS[2]

        # for no UniProt in SIFTS
        if len(tuple_PDBe_for_UniProt_and_tuple_UniProt) == 0:
            copy_file(default_input_path_to_PDB, PDB, default_output_path_to_PDB, "_renum.pdb" + assembly_num, gzip_mode)
            log_message = if_no_SIFTS_data_log_for_PDB(default_input_path_to_PDB, PDB_id, PDB)
            return log_message

        split = try_PDB(default_input_path_to_PDB, PDB)
        if split == 0:
            continue

        product_of_SIFTS_data_parser = SIFTS_data_parser_for_PDB(tuple_PDBe_for_PDB_and_tuple_PDB,
                                                                 tuple_PDBe_for_UniProt_and_tuple_UniProt,
                                                                 default_PDB_num, 'all')
        df_PDBe_PDB_UniProt = product_of_SIFTS_data_parser[1]

        handling_chain_numbering = handling_chain_numbering_clashes(df_PDBe_PDB_UniProt, exception_AccessionIDs)
        chains_to_change = handling_chain_numbering[0]
        combined_tuple_PDBe_UniProt_AccessionID = handling_chain_numbering[1]
        longest_AccessionID_list = handling_chain_numbering[3]
        chains_to_change_one_to_end = handling_chain_numbering[4]

        product_of_SIFTS_data_parser = SIFTS_data_parser_for_PDB(tuple_PDBe_for_PDB_and_tuple_PDB, combined_tuple_PDBe_UniProt_AccessionID,
                                                                 default_PDB_num, chains_to_change)
        df_PDBe_PDB_UniProt_without_null_index_PDBe = product_of_SIFTS_data_parser[0]
        df_PDBe_PDB_UniProt = product_of_SIFTS_data_parser[1]

        renumbered_count = renumbered_count_in_chains(chains_to_change_one_to_end, df_PDBe_PDB_UniProt_without_null_index_PDBe,
                                                      PDB_id, UniProt_conversion_dict, longest_AccessionID_list)
        chain_total_renum = renumbered_count[0]
        nothing_changed = renumbered_count[1]

        chain_total_renum.append(nothing_changed)
        mod_log_message = chain_total_renum

        if nothing_changed == 0:
            copy_file(default_input_path_to_PDB, PDB, default_output_path_to_PDB, "_renum.pdb" + assembly_num, gzip_mode)
            return mod_log_message

        parsed_PDB = PDB_parser(split, df_PDBe_PDB_UniProt_without_null_index_PDBe, default_PDB_num)
        pd_series_index_PDBe = parsed_PDB[0]
        num_ins_code_name_chain_HETATM = parsed_PDB[1]
        df_final_poly_corrected = parsed_PDB[2]
        missing_res_remark_465 = parsed_PDB[3]
        Num_is_Too_Big = parsed_PDB[4]

        # when numbers get too big
        if not Num_is_Too_Big:
            copy_file(default_input_path_to_PDB, PDB, default_output_path_to_PDB, "_renum.pdb" + assembly_num, gzip_mode)
            return mod_log_message

        df_final_nonpoly_corrected = non_poly_num(pd_series_index_PDBe, num_ins_code_name_chain_HETATM)
        if len(missing_res_remark_465) != 0:
            final_remark_465 = remark_465(missing_res_remark_465, df_PDBe_PDB_UniProt)
        else:
            final_remark_465 = None

        dict_for_replacement = final_dict_formation(df_final_poly_corrected, df_final_nonpoly_corrected, final_remark_465, chains_to_change)

        outF = open(Path(str(default_output_path_to_PDB) + "/" + PDB_id + ".pdb" + assembly_num), "w")
        # PDBrenum REMARK 0 insert
        start_remark_0 = 0
        for lines in split:
            if lines.startswith("HEADER"):
                start_remark_0 += 1
            else:
                split = split[:start_remark_0] + PDBrenum_REMARK + split[start_remark_0:]
                break

        # actual renumbering
        for lines in split:
            lines = replace_all(lines, dict_for_replacement)
            outF.write(lines)
            outF.write("\n")
        outF.close()

        if gzip_mode == "on":
            with open(Path(str(default_output_path_to_PDB) + "/" + PDB_id + ".pdb" + assembly_num), 'rb') as f_in:
                with gzip.open(Path(str(default_output_path_to_PDB) + "/" + PDB_id + "_renum.pdb" + assembly_num + ".gz"), 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)
            os.remove(Path(str(default_output_path_to_PDB) + "/" + PDB_id + ".pdb" + assembly_num))

        return mod_log_message


def ProcessPool_run_renum_PDB(format_to_download, input_PDB_files_were_found, default_input_path_to_PDB, default_input_path_to_SIFTS,
                              default_output_path_to_PDB, default_PDB_num, gzip_mode, exception_AccessionIDs, nproc):
    if not os.path.exists(default_output_path_to_PDB):
        os.makedirs(default_output_path_to_PDB)

    resulting = list()
    executor = ProcessPoolExecutor(max_workers=nproc)
    partial_master_PDB_renumber_function = partial(master_PDB_renumber_function,
                                                   default_input_path_to_PDB=default_input_path_to_PDB,
                                                   default_input_path_to_SIFTS=default_input_path_to_SIFTS,
                                                   default_output_path_to_PDB=default_output_path_to_PDB,
                                                   default_PDB_num=default_PDB_num, gzip_mode=gzip_mode,
                                                   exception_AccessionIDs=exception_AccessionIDs)

    jobs = [executor.submit(partial_master_PDB_renumber_function, pdb_files) for pdb_files in input_PDB_files_were_found]

    for job in tqdm.tqdm(as_completed(jobs), total=len(jobs), position=0, miniters=1,
                         leave=True, desc="Renumbering " + format_to_download + " files"):
        result = job.result()
        resulting.append(result)

    return resulting


# from src.download.lookfilesinside import look_what_is_inside
# current_directory = "/home/bulat/Desktop/main/PDB_fix_project/PDBarnum"
# default_input_path_to_mmCIF = current_directory + "/mmCIF"
# default_input_path_to_PDB = current_directory + "/PDB"
# default_input_path_to_SIFTS = current_directory + "/SIFTS"
# default_output_path_to_mmCIF = current_directory + "/output_mmCIF"
# default_output_path_to_PDB = current_directory + "/output_PDB"
#
# default_input_path_to_mmCIF_assemblies = current_directory + "/mmCIF_assembly"
# default_input_path_to_PDB_assemblies = current_directory + "/PDB_assembly"
# default_output_path_to_mmCIF_assemblies = current_directory + "/output_mmCIF_assembly"
# default_output_path_to_PDB_assemblies = current_directory + "/output_PDB_assembly"
#
# #
# # current_directory = os.getcwd()
# # default_mmCIF_num = 50000
# default_PDB_num = 5000
# gzip_mode = "on"
# exception_AccessionIDs = ["P42212", "Q17104", "Q27903", "Q93125", "P03069", "D3DLN9", "Q96UT3", "P0ABE7", "P00192",
#                           "P76805", "Q8XCE3", "P00720", "Q38170", "Q94N07", "P0AEX9", "P02928", "Q2M6S0"]

# input_PDB_files_were_found = look_what_is_inside("PDB", default_input_path_to_PDB=default_input_path_to_PDB)

# if __name__ == '__main__':
#     resulting2 = ProcessPool_run_renum_PDB("PDB", input_PDB_files_were_found,
#                                            default_input_path_to_PDB,
#                                            default_input_path_to_SIFTS,
#                                            default_output_path_to_PDB,
#                                            default_PDB_num, gzip_mode, exception_AccessionIDs)



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import imp
import warnings

warnings.filterwarnings("ignore")

import pandas as pd
from tqdm import tqdm
import concurrent.futures

from ..functions.lst import lst_to_str, str_to_lst, res_to_lst
from ..functions.coord import build_add_resid_lst, resname_to_letter
from ..functions.col import (
    core_path_col,
    modelid_col,
    chainid_col,
    phi_col,
    psi_col,
    omega_col,
    rama_col,
    rotamer_col,
    sc_col_lst,
    complete_col,
    resname_col,
    dih_col_lst,
    bb_seq_col,
    bb_len_col,
    bb_resid_col,
    sc_len_col,
    sc_resid_col,
    sc_seq_col,
)
from ..functions.dih import get_rama_type, get_rot_type
from ..functions.table import (
    get_df_at_index,
    fix_val,
    get_val_col,
    get_col_val_lst,
    get_val_col_lst,
)
from ..functions.path import save_table


def get_index_dih(
    df, index, dih_dict, resid_dict, max_ca_dist=4, ext_mult=1, coord_path_col=None
):

    if coord_path_col is None:
        coord_path_col = core_path_col

    index_df = get_df_at_index(df, index)

    coord_path = index_df.at[index, coord_path_col]

    if coord_path in list(dih_dict.keys()):

        modelid = fix_val(index_df.at[index, modelid_col], return_int=True)
        chainid = index_df.at[index, chainid_col]

        try:
            dih_resid_lst = list(dih_dict[coord_path][str(modelid)][chainid].keys())
        except:
            dih_resid_lst = list(dih_dict[coord_path][int(modelid)][chainid].keys())

        index_df.at[index, complete_col] = True

        bb_resid_lst = list()
        bb_seq_lst = list()

        sc_seq_lst = list()
        sc_resid_lst = list()
        sc_len_lst = list()

        for dih_col in dih_col_lst:

            resid_lst = resid_dict[dih_col]

            if resid_lst is not None:

                add_resid_lst = build_add_resid_lst(
                    coord_path,
                    modelid,
                    chainid,
                    resid_lst,
                    dih_resid_lst,
                    max_ca_dist=max_ca_dist,
                    ext_mult=ext_mult,
                )

                val = 999.00
                resname = "-"
                for i, add_resid in enumerate(add_resid_lst):

                    if str(add_resid) in dih_resid_lst:

                        try:
                            val = dih_dict[coord_path][str(modelid)][chainid][
                                str(add_resid)
                            ][dih_col]
                            resname = resname_to_letter(
                                dih_dict[coord_path][str(modelid)][chainid][
                                    str(add_resid)
                                ][resname_col]
                            )
                        except:
                            val = dih_dict[coord_path][int(modelid)][chainid][
                                str(add_resid)
                            ][dih_col]
                            resname = resname_to_letter(
                                dih_dict[coord_path][int(modelid)][chainid][
                                    str(add_resid)
                                ][resname_col]
                            )

                    if dih_col == phi_col:
                        bb_resid_lst.append(add_resid)
                        bb_seq_lst.append(resname)

                    if dih_col in sc_col_lst:
                        sc_resid_lst.append(add_resid)
                        sc_seq_lst.append(resname)

                    val_col = get_val_col(dih_col, i + 1)

                    index_df.at[index, val_col] = float(val)

                    if val == 999.00:
                        index_df.at[index, complete_col] = False

                    val = 999.00
                    resname = "-"

                if dih_col in sc_col_lst:
                    sc_len_lst.append(len(sc_resid_lst))

        index_df.at[index, bb_len_col] = len(bb_resid_lst)
        index_df.at[index, sc_len_col] = lst_to_str(sc_len_lst)
        index_df.at[index, bb_resid_col] = lst_to_str(bb_resid_lst)
        index_df.at[index, sc_resid_col] = lst_to_str(sc_resid_lst)
        index_df.at[index, bb_seq_col] = lst_to_str(bb_seq_lst, join_txt="")
        index_df.at[index, sc_seq_col] = lst_to_str(sc_seq_lst, join_txt="")

    else:
        index_df = pd.DataFrame()

    return index_df


def get_rama_str(bb_vals_lst):

    rama_str = ""

    for bb_vals in bb_vals_lst:

        if 999.00 in bb_vals:
            rama_type = "-"
        else:
            rama_type = get_rama_type(bb_vals[0], bb_vals[1], bb_vals[2])

        rama_str += rama_type

    return rama_str


def get_rot_str(rot_val_lst):

    rot_str = ""

    for rot_val in rot_val_lst:

        rot_str += get_rot_type(rot_val)

    return rot_str


def add_bb_rama(df):

    bb_resid_lst = get_col_val_lst(df, phi_col)

    if len(bb_resid_lst) > 0:

        index_lst = list(df.index.values)

        for index in index_lst:

            max_index = int(float(df.at[index, bb_len_col]))

            if max_index > 1:
                resid_lst = bb_resid_lst[:max_index]
            else:
                resid_lst = bb_resid_lst

            bb_val_lst = list()

            for resid in resid_lst[:max_index]:

                bb_val_lst.append(
                    (
                        df.at[index, get_val_col(phi_col, resid)],
                        df.at[index, get_val_col(psi_col, resid)],
                        df.at[index, get_val_col(omega_col, resid)],
                    )
                )

            df.at[index, rama_col] = get_rama_str(bb_val_lst)

    return df


def add_sc_rot(df):

    for index in list(df.index.values):

        if sc_len_col in list(df.columns):

            sc_len_lst = str_to_lst(df.at[index, sc_len_col])

            sc_val_lst = list()
            i = 0

            for sc_col in sc_col_lst:

                sc_resid_lst = get_col_val_lst(df, sc_col)

                if len(sc_resid_lst) > 0:

                    max_index = int(float(sc_len_lst[i]))

                    if max_index > 1:
                        resid_lst = sc_resid_lst[:max_index]
                    else:
                        resid_lst = sc_resid_lst

                    for resid in resid_lst:

                        sc_val_lst.append(df.at[index, get_val_col(sc_col, resid)])

                    i += 1

            if len(sc_val_lst) > 0:

                df.at[index, rotamer_col] = get_rot_str(sc_val_lst)

    return df


def build_dih_table(
    df,
    dih_dict,
    dih_table_path=None,
    bb_resids=None,
    chi1_resids=None,
    chi2_resids=None,
    altchi1_resids=None,
    altchi2_resids=None,
    chi3_resids=None,
    chi4_resids=None,
    chi5_resids=None,
    max_ca_dist=4,
    ext_mult=1,
    coord_path_col=None,
    num_cpu=1,
):
    df = df.reset_index(drop=True)

    df_col_lst = list(df.columns)

    check_col_lst = [
        bb_resid_col,
        sc_resid_col,
        bb_seq_col,
        sc_seq_col,
        bb_len_col,
        sc_len_col,
        rama_col,
        rotamer_col,
    ]

    for check_col in check_col_lst:
        if check_col in df_col_lst:
            del df[check_col]

    for dih_col in dih_col_lst:
        for val_col in get_val_col_lst(df, dih_col):
            del df[val_col]

    angle_resids_lst = [
        bb_resids,
        bb_resids,
        bb_resids,
        chi1_resids,
        chi2_resids,
        altchi1_resids,
        altchi2_resids,
        chi3_resids,
        chi4_resids,
        chi5_resids,
    ]

    resid_dict = dict()

    for index, angle_resids in enumerate(angle_resids_lst):

        resid_dict[dih_col_lst[index]] = res_to_lst(angle_resids)

    dih_df = pd.DataFrame()

    if num_cpu == 1:
        for index in tqdm(
            list(df.index.values),
            desc="Building dihedral table",
            position=0,
            leave=True,
        ):
            index_df = get_index_dih(
                df,
                index,
                dih_dict,
                resid_dict,
                max_ca_dist=max_ca_dist,
                ext_mult=ext_mult,
                coord_path_col=coord_path_col,
            )
            for col in list(index_df.columns):
                if col not in list(dih_df.columns):
                    dih_df[col] = 999.00

            dih_df = pd.concat([dih_df, index_df], sort=False)
    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:
            job_lst = [
                executor.submit(
                    get_index_dih,
                    df,
                    index,
                    dih_dict,
                    resid_dict,
                    max_ca_dist=max_ca_dist,
                    ext_mult=ext_mult,
                    coord_path_col=coord_path_col,
                )
                for index in list(df.index.values)
            ]

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Building dihedral table",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):
                index_df = job.result()

                for col in list(index_df.columns):
                    if col not in list(dih_df.columns):
                        dih_df[col] = 999.00

                dih_df = pd.concat([dih_df, index_df], sort=False)

    dih_df = dih_df.reset_index(drop=True)

    for dih_col in dih_col_lst:
        for val_col in get_val_col_lst(dih_df, dih_col):
            dih_df[val_col] = dih_df[val_col].fillna(999.00)

    dih_df = add_bb_rama(dih_df)
    dih_df = add_sc_rot(dih_df)

    print("Built dihedral table!")

    if dih_table_path is not None:
        save_table(dih_table_path, dih_df)
    else:
        return dih_df


# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from tqdm import tqdm
import concurrent.futures

from ..functions.path import save_json
from ..functions.table import merge_dicts
from ..functions.dih import calc_bb_angle, calc_sc_angle
from ..functions.coord import (
    load_coord,
    get_modelid,
    get_chainid,
    get_resid,
    get_resname,
    get_resid_str,
    is_aa,
)
from ..functions.col import resname_col, bb_col_lst, sc_col_lst
from ..functions.lst import type_lst
from ..functions.table import merge_dicts
from ..functions.path import save_json


def build_dih_dict(coord_path):

    structure = load_coord(coord_path)

    dih_dict = {coord_path: dict()}

    for model in structure:
        modelid = get_modelid(model)
        dih_dict[coord_path][modelid] = dict()
        for chain in model:
            chainid = get_chainid(chain)

            dih_dict[coord_path][modelid][chainid] = dict()

            resid_lst = [get_resid(residue) for residue in chain if is_aa(residue)]
            resname_lst = [get_resname(residue) for residue in chain if is_aa(residue)]
            resstr_lst = [get_resid_str(residue) for residue in chain if is_aa(residue)]

            for index, resid in enumerate(resid_lst):

                curr_resid = resid
                curr_resname = resname_lst[index]
                curr_resid_str = resstr_lst[index]

                dih_dict[coord_path][modelid][chainid][curr_resid_str] = dict()
                dih_dict[coord_path][modelid][chainid][curr_resid_str][
                    resname_col
                ] = curr_resname

                if index != 0 and index != len(resid_lst) - 1:
                    prev_resid = resid_lst[index - 1]
                    next_resid = resid_lst[index + 1]

                    for bb_col in bb_col_lst:

                        bb_val = calc_bb_angle(
                            structure=structure,
                            modelid=modelid,
                            chainid=chainid,
                            curr_resid=curr_resid,
                            prev_resid=prev_resid,
                            next_resid=next_resid,
                            angle=bb_col,
                        )

                        dih_dict[coord_path][modelid][chainid][curr_resid_str][
                            bb_col
                        ] = bb_val

                else:
                    for bb_col in bb_col_lst:
                        dih_dict[coord_path][modelid][chainid][curr_resid_str][
                            bb_col
                        ] = 999.00

                for sc_col in sc_col_lst:

                    sc_val = calc_sc_angle(
                        structure=structure,
                        modelid=modelid,
                        chainid=chainid,
                        resid=curr_resid,
                        angle=sc_col,
                    )
                    dih_dict[coord_path][modelid][chainid][curr_resid_str][
                        sc_col
                    ] = sc_val

    return dih_dict


def prep_dih(coord_paths, dih_json_path=None, num_cpu=1):

    coord_path_lst = type_lst(coord_paths)

    dih_dict = dict()

    if num_cpu == 1:
        for coord_path in tqdm(
            coord_path_lst,
            desc="Preparing dihedrals",
            position=0,
            leave=True,
        ):
            dih_dict = merge_dicts([dih_dict, build_dih_dict(coord_path)])

    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:
            job_lst = [
                executor.submit(build_dih_dict, coord_path)
                for coord_path in coord_path_lst
            ]

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Preparing dihedrals",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):
                dih_dict = merge_dicts([dih_dict, job.result()])

    print("Prepared dihedrals!")

    if dih_json_path is not None:
        save_json(dih_json_path, dih_dict)
    else:
        return dih_dict



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd

from ..scripts.build_dih_table import build_dih_table
from ..scripts.make_facet_plot import make_facet_plot
from ..functions.lig import lig_lst_dict
from ..functions.color import red_hex, gray_hex
from ..functions.table import ( build_col_count_dict, mask_equal, lst_col,
                            extract_int, fix_col, make_dict)
from ..functions.col import (pocket_class_col, pdb_id_col, pdb_code_col,
                            chainid_col, bound_prot_swiss_id_col, pharm_class_col, bound_prot_chainid_col,
                            nuc_class_col, mut_status_col, prot_class_col, interf_class_col,
                            bio_lig_col, ion_lig_col, mem_lig_col, pharm_lig_col, match_class_col, 
                            modelid_col, gene_class_col, prot_class_col,
                            phi_col, psi_col, chi1_col, chi2_col,
                            pharm_lig_col, rename_col_dict)
from ..functions.file import entry_table_file, dih_json_file
from ..functions.path import rascore_str
from ..constants.pharm import pharm_color_dict, sp2_name, other_pharm_name, none_pharm_name
from ..constants.prot import prot_color_dict, none_prot_name, other_prot_name
from ..constants.conf import loop_resid_dict, loop_color_dict, sw1_name, sw2_name, y32_name, y71_name, sw1_color, sw2_color
from ..functions.gui import (load_st_table, show_st_dataframe, write_st_end,
                            show_st_structure, show_st_table, get_html_text, 
                            download_st_df, rename_st_cols, reorder_st_cols)
from ..functions.lst import res_to_lst, lst_nums, str_to_lst

import streamlit as st

reverse_col_dict = make_dict(list(rename_col_dict.values()),list(rename_col_dict.keys()))

sw1_resid_lst = res_to_lst(loop_resid_dict[sw1_name])
sw2_resid_lst = res_to_lst(loop_resid_dict[sw2_name])

def mutation_page():
    
    st.markdown("## Analyze Mutations")

    st.markdown("---")

    df = load_st_table(__file__)

    st.sidebar.markdown("""**Note.** We recommend comparing mutated structures with 
    identical SW1 and SW2 conformations and molecular annotations.
    """)

    left_name = "Left"
    right_name = "Right"

    a_name = "G12D"
    b_name = "G12V"

    left_color = red_hex
    right_color = gray_hex

    left_query_col, right_query_col = st.columns(2)

    mut_status_lst = lst_col(df, mut_status_col, unique=True)

    left_name = left_query_col.selectbox(f"{rename_col_dict[mut_status_col]} ({left_name})", mut_status_lst, index=mut_status_lst.index(a_name))

    mut_status_lst = [x for x in mut_status_lst if x != left_name]

    right_name = right_query_col.selectbox(f"{rename_col_dict[mut_status_col]} ({right_name})", mut_status_lst, index=mut_status_lst.index(b_name))

    query_col_dict = {left_name: left_query_col, right_name: right_query_col}

    query_df_dict = dict()

    for query_name, query_col in query_col_dict.items():

        query_df = df.copy(deep=True)

        query_df = mask_equal(query_df, mut_status_col, query_name)

        if query_col.checkbox(f"Display Selection Options ({query_name})", value=False):

            query_col.markdown(f"#### Conformation Selection ({query_name})")

            for col in [sw1_name, sw2_name, y32_name, y71_name]:

                mask_lst = query_col.multiselect(f"{rename_col_dict[col]} ({query_name})", lst_col(query_df, col, unique=True))

                if len(mask_lst) > 0:
                    query_df = mask_equal(query_df, col, mask_lst)

            query_col.markdown(f"#### Annotation Selection ({query_name})")

            for col in [gene_class_col, nuc_class_col, prot_class_col, pocket_class_col, match_class_col, interf_class_col]:

                mask_lst = query_col.multiselect(f"{rename_col_dict[col]} ({query_name})", lst_col(query_df, col, unique=True))

                if len(mask_lst) > 0:
                    query_df = mask_equal(query_df, col, mask_lst)

        query_df_dict[query_name] = query_df

    mut_df = pd.concat([query_df_dict[left_name], query_df_dict[right_name]], sort=False)

    with st.expander("One-to-One Comparison", expanded=False):

        st.markdown("#### One-to-One")

        left_info_col, right_info_col = st.columns(2)   

        info_col_dict = {left_name: left_info_col, right_name: right_info_col}   

        pdb_name_dict = dict()
        pdb_df_dict = dict()

        for query_name, query_df in query_df_dict.items():

            pdb_id_lst = [x.upper() for x in lst_col(query_df, pdb_id_col) if x.upper() not in list(pdb_name_dict.values())]

            info_col = info_col_dict[query_name]
            
            pdb_upper = info_col.selectbox(f"PDB ID ({query_name})", pdb_id_lst)

            pdb_code = pdb_upper[:4].lower()
            chainid = pdb_upper[4:5]

            pdb_id = f"{pdb_code}{chainid}"
             
            pdb_df = mask_equal(query_df, pdb_id_col, pdb_id)
                
            pdb_name_dict[query_name] = pdb_upper
            pdb_df_dict[query_name] = pdb_df

            pdb_code = pdb_df[pdb_code_col].iloc[0]
            chainid = pdb_df[chainid_col].iloc[0]
            gene_class = pdb_df[gene_class_col].iloc[0]      

            mut_status = pdb_df[mut_status_col].iloc[0]

            info_col.markdown(f"##### PDB: [{pdb_code.upper()}](https://www.rcsb.org/structure/{pdb_code}) (Chain {chainid}) - {gene_class}({mut_status})")

            col_pair_dict = {nuc_class_col: bio_lig_col, match_class_col: pharm_lig_col, prot_class_col: bound_prot_swiss_id_col}
            for col_1, col_2 in col_pair_dict.items():
                val_1 = pdb_df[col_1].iloc[0]
                val_2 = pdb_df[col_2].iloc[0]
                col_str = f"**{rename_col_dict[col_1]}:** {val_1}"
                if val_1 != "None":
                    if col_2 in [bio_lig_col, pharm_lig_col] and val_2 != "None":
                        col_str += f" ([{val_2}](https://www.rcsb.org/ligand/{val_2}))"
                    elif col_2 in [bound_prot_swiss_id_col] and val_2 != "None":
                        col_str += f" ([{val_2}](https://www.uniprot.org/uniprot/{val_2}))"
                    else:
                        col_str += f" ({val_2})"
                info_col.markdown(col_str)
           
        st.markdown("---")

        st.markdown("##### Viewer Settings") 

        left_set_col, right_set_col = st.columns(2)  

        left_color = left_set_col.color_picker(label=f"{left_name} Color", value=left_color)
        right_color = left_set_col.color_picker(label=f"{right_name} Color", value=right_color)

        stick_resid_lst = left_set_col.multiselect("Displayed Residues (Mutations Always Shown)", lst_nums(1, 189), default=[12, 13, 32, 61])

        label_muts = left_set_col.checkbox("Label Mutations",value=True)
        label_resids = left_set_col.checkbox("Label Residues",value=False)

        style_dict = {"Ribbon": "oval", "Trace": "trace"}

        cartoon_style = style_dict[
            right_set_col.radio("Cartoon Style", ["Ribbon", "Trace"])
        ]

        cartoon_trans = right_set_col.slider(
            "Cartoon Transparency", min_value=0.0, max_value=1.0, value=0.5
        )

        surface_trans = right_set_col.slider(
            "Surface Transparency (Mutations Only)", min_value=0.0, max_value=1.0, value=0.5
        )

        left_view_col, right_view_col = st.columns(2)   

        view_col_dict = {left_name: left_view_col, right_name: right_view_col}  

        for query_name, query_df in query_df_dict.items():
            
            pdb_df = pdb_df_dict[query_name] 

            pdb_code = pdb_df[pdb_code_col].iloc[0]
            chainid = pdb_df[chainid_col].iloc[0]

            mut_status = pdb_df[mut_status_col].iloc[0]
            
            mut_resid_lst = list()
            if mut_status != "WT":
                mut_resid_lst = [extract_int(x) for x in str_to_lst(mut_status)]

            if mut_status == left_name:
                mut_color = left_color
            elif mut_status == right_name:
                mut_color = right_color

            view_col = view_col_dict[query_name]

            show_st_structure(pdb_df,
                    mut_resids=mut_resid_lst,
                    stick_resids=stick_resid_lst,
                    label_resids=label_resids, 
                    label_muts=label_muts,
                    mut_color=mut_color,
                    cartoon_style=cartoon_style,
                    cartoon_trans=cartoon_trans, 
                    mut_trans=surface_trans,
                    zoom_resids=mut_resid_lst + stick_resid_lst,
                    zoom=0.7,
                    width=400,
                    height=400,
                    st_col=view_col)
            
            label_size = "medium"

            for col in [sw1_name, sw2_name, y32_name, y71_name]:

                label_str = get_html_text({f"{rename_col_dict[col]}: ": "#31333F"}, font_weight='bold', font_size=label_size)

                if col in [y32_name, sw1_name]:
                    label_color = loop_color_dict[sw1_name]
                elif col in [y71_name, sw2_name]:
                    label_color = loop_color_dict[sw2_name]

                label_str += get_html_text({pdb_df[col].iloc[0]: label_color}, font_size=label_size)

                view_col.markdown(label_str, unsafe_allow_html=True)

    with st.expander("Many-to-Many Comparison", expanded=False):

        st.markdown("#### Many-to-Many")

        count_dict = build_col_count_dict(df, mut_status_col)

        left_plot_col, middle_plot_col, right_plot_col = st.columns(3)

        left_color = left_plot_col.color_picker(label=f"{left_name} (N={count_dict[left_name]})", value=left_color)
        right_color = left_plot_col.color_picker(label=f"{right_name} (N={count_dict[right_name]})", value=right_color)

        marker_size = left_plot_col.number_input("Marker Size", min_value=1, max_value=50, value=5)

        bb_name = "Backbone Dihedrals (Phi vs. Psi)"
        sc_name = "Side Chain Dihedrals (Chi1 vs. Chi2)"

        plot_type = middle_plot_col.radio("Plot Type", [bb_name, sc_name])

        dih_resids = middle_plot_col.selectbox("Dihedral Residue", lst_nums(1,166), index=31)

        bb_resids = None
        chi1_resids = None
        chi2_resids = None

        if plot_type == bb_name:
            bb_resids = dih_resids
            x_col = phi_col
            y_col = psi_col
        elif plot_type == sc_name:
            chi1_resids = dih_resids
            chi2_resids = dih_resids
            x_col = chi1_col
            y_col = chi2_col

        if middle_plot_col.button("Plot Dihedrals"):

            with st.spinner("Preparing Dihedrals"):

                dih_dict = load_st_table(__file__, file_name=dih_json_file, json_format=True)

                mut_df = build_dih_table(mut_df, dih_dict, bb_resids=bb_resids, 
                                            chi1_resids=chi1_resids,
                                            chi2_resids=chi2_resids)

            fig = make_facet_plot(
                        mut_df,
                        x_col=x_col,
                        y_col=y_col,
                        show_legend=False,
                        hue_col=mut_status_col,
                        hue_order=[left_name, right_name],
                        hue_palette={left_name: left_color, right_name: right_color},
                        plot_width=3,
                        plot_height=3,
                        font_size=12,
                        marker_size=marker_size,
                    )

            right_plot_col.pyplot(fig)

        st.markdown("---")

        left_conf_col, right_conf_col = st.columns(2)

        for table_col in [y32_name, y71_name, sw1_name, sw2_name]:

            if table_col in [y32_name, sw1_name]:
                conf_col = left_conf_col
            elif table_col in [y71_name, sw2_name]:
                conf_col = right_conf_col

            conf_col.markdown(f"##### {rename_col_dict[table_col]}")

            loop_df = (
                pd.pivot_table(
                    data=rename_st_cols(mut_df),
                    index=[rename_col_dict[table_col]],
                    columns=rename_col_dict[mut_status_col],
                    values=rename_col_dict[pdb_id_col],
                    aggfunc="nunique",
                    margins=True,
                )
                .fillna("")
            )

            for col in list(loop_df.columns):
                loop_df[col] = loop_df[col].map(str)
                loop_df = fix_col(loop_df, col)

            loop_df = reorder_st_cols(loop_df, table_col, mut_status_col)

            loop_df = loop_df.reset_index()

            show_st_table(loop_df, st_col=conf_col)

        #     st.markdown("---")

        #     left_site_col, right_chem_col = st.columns(2)

        #     pdb_1 = pdb_name_dict[left_name]
        #     pdb_2 = pdb_name_dict[right_name]

        #     pdb_1_df = pdb_df_dict[left_name]
        #     pdb_2_df = pdb_df_dict[right_name]

        #     lig_1 = pdb_1_df[pharm_lig_col].iloc[0]
        #     lig_2 = pdb_2_df[pharm_lig_col].iloc[0]

        #     smiles_dict_1 = str_to_dict(pdb_1_df[pharm_lig_smiles_col].iloc[0])
        #     smiles_dict_2 = str_to_dict(pdb_2_df[pharm_lig_smiles_col].iloc[0])
            
        #     smiles_1 = smiles_dict_1[lig_1][0]
        #     smiles_2 = smiles_dict_2[lig_2][0]

        #     right_chem_col.markdown("#### Chemistry Comparison")

        #     mcs = get_lig_mcs([smiles_1, smiles_2])

        #     right_chem_col.markdown(f"**2D Fingerprint Similarity:** {round(get_lig_simi(smiles_1, smiles_2),2)}")

        #     right_chem_col.markdown(f"**Maximum Common Substructure (MCS):** {get_lig_smiles(mcs)}")

        #     if not right_chem_col.checkbox("Highlight MCS",value=True):
        #         mcs = None
            
        #     right_chem_col.image(draw_lig_plot([smiles_1, smiles_2], 
        #     lig_labels=[f"{pdb_1} ({lig_1})", f"{pdb_2} ({lig_2})"], font_size=16, mol_pad=0,
        #     plot_height=4, plot_width=3, highlight_querys=mcs, color_palette=[pharm_color_dict[pharm_class]]))

        #     cont_dict_1 = str_to_dict(pdb_1_df[bound_lig_cont_col].iloc[0])
        #     cont_dict_2 = str_to_dict(pdb_2_df[bound_lig_cont_col].iloc[0])

        #     cont_lst_1 = cont_dict_1[lig_1]
        #     cont_lst_2 = cont_dict_2[lig_2]
            
        #     cont_df = pd.DataFrame()

        #     left_site_col.markdown("#### Site Comparison")

        #     left_site_col.markdown(f"**Residue Contact Similarity (Simpson):** {round(calc_simpson(cont_lst_1, cont_lst_2),2)}")

        #     cont_col_1 = f"{pdb_1} ({lig_1})"
        #     cont_col_2 = f"{pdb_2} ({lig_2})"

        #     for resid in sorted(lst_unique(cont_lst_1 + cont_lst_2,return_int=True)):

        #         cont_1 = ''
        #         cont_2 = ''

        #         if str(resid) in cont_lst_1:
        #             cont_1 = "✓"
        #         if str(resid) in cont_lst_2:
        #             cont_2 = "✓"

        #         cont_df.at[resid, cont_col_1] = cont_1
        #         cont_df.at[resid, cont_col_2] = cont_2

        #     cont_df = cont_df.rename_axis('Residue Contact').reset_index()

        #     show_st_table(cont_df, st_col=left_site_col)


        # with st.expander("Many-to-Many Comparison", expanded=False):

        #     st.markdown("#### Many-to-Many Comparison")

        #     count_dict = build_col_count_dict(many_df, id_col)

        #     color_dict = {left_name: pharm_color_dict[pharm_class], 
        #     right_name: get_hex(change_hex_alpha(pharm_color_dict[pharm_class], 0.5)), both_name: gray_hex}

        #     left_plot_col, middle_plot_col, right_plot_col = st.columns(3)

        #     for label_name, label_color in color_dict.items():

        #         count = 0
        #         if label_name in list(count_dict.keys()):
        #             count = count_dict[label_name]

        #         color_dict[label_name] = left_plot_col.color_picker(label=f"{label_name} (N={count})", 
        #                                                         value=label_color)

        #     many_order =  [x for x in list(color_dict.keys()) if x in lst_col(many_df, id_col, unique=True)]

        #     scatter_name = "Scatterplot"
        #     box_name = "Boxplot"

        #     plot_type = middle_plot_col.radio("Plot Type", [scatter_name, box_name])

        #     marker_size = middle_plot_col.number_input("Marker Size", min_value=1, max_value=50, value=5)
            
        #     if plot_type == scatter_name:
        #         plot_reg = middle_plot_col.checkbox("Display Regression",value=True)

        #         fig = make_facet_plot(many_df, 
        #             x_col=pocket_volume_col,
        #             y_col=pocket_score_col,
        #             x_str=rename_col_dict[pocket_volume_col],
        #             y_str=rename_col_dict[pocket_score_col],
        #             x_round=0,
        #             y_round=1,
        #             hue_col=id_col,
        #             hue_palette=color_dict,
        #             hue_order=many_order,
        #             plot_width=3,
        #             plot_height=3,
        #             font_size=12,
        #             x_rotation=45,
        #             marker_size=marker_size,
        #             x_pad=25,
        #             y_pad=10,
        #             x_ha="right",
        #             plot_reg=plot_reg,
        #             log_reg=True,
        #             trun_reg=True,
        #             x_lim=[0, 1700],
        #             x_ticks=[0, 500, 1000, 1500],
        #             y_lim=[0, 1.2],
        #             y_ticks=[0.0, 0.5, 1.0])
            
        #     elif plot_type == box_name:

        #         y_str = middle_plot_col.radio("Y-Axis",[rename_col_dict[pocket_score_col],rename_col_dict[pocket_volume_col]])

        #         y_col = reverse_col_dict[y_str]

        #         if y_col == pocket_score_col:
        #             y_lim = [0, 1.2]
        #             y_ticks = [0.0, 0.5, 1.0]
        #             y_round = 1

        #         elif y_col == pocket_volume_col:
        #             y_lim = [0, 1700]
        #             y_ticks = [0, 500, 1000, 1500]
        #             y_round = 0

        #         stat_pairs = None
        #         if len(many_order) > 1:
        #             if middle_plot_col.checkbox("Display Welch's t-test p-value",value=True):
        #                 stat_pairs = list()
        #                 for id_1 in many_order:
        #                     for id_2 in many_order:
        #                         if id_1 != id_2:
        #                             if (id_2, id_1) not in stat_pairs:
        #                                 stat_pairs.append((id_1, id_2))

        #         fig = make_facet_plot(
        #                 many_df,
        #                 x_col=id_col,
        #                 x_order=many_order,
        #                 x_palette=color_dict,
        #                 x_count=True,
        #                 x_str="",
        #                 y_col=y_col,
        #                 y_str=y_str,
        #                 show_legend=False,
        #                 plot_width=3,
        #                 plot_height=3,
        #                 plot_kind="box",
        #                 font_size=12,
        #                 marker_size=marker_size,
        #                 x_pad=35,
        #                 y_lim=y_lim,
        #                 y_ticks=y_ticks,
        #                 y_round=y_round,
        #                 stat_pairs=stat_pairs,
        #                 stat_test="t-test_welch",
        #                 stat_format="simple"
                
                    
        #             )

        #     right_plot_col.pyplot(fig)

        #     many_df[pocket_score_col] = many_df[pocket_score_col].map(float)
        #     many_df[pocket_volume_col] = many_df[pocket_volume_col].map(float)

        #     left_stat_col, right_stat_col = st.columns(2)

        #     stat_col_dict = {pocket_score_col: left_stat_col, pocket_volume_col: right_stat_col}

        #     for stat_name, stat_col in stat_col_dict.items():

        #         stat_col.markdown(f"##### {rename_col_dict[stat_name]}")

        #         temp_df = pd.pivot_table(many_df, values=stat_name,
        #                                 index=id_col,
        #                                 aggfunc=[np.mean, min, max])

        #         stat_df = pd.DataFrame()
        #         for index in list(temp_df.index.values):
        #             for col, stat in zip(list(temp_df.columns), [mean_name, min_name, max_name]):
        #                 stat_df.at[index, stat] = temp_df.at[index, col]

        #         for col in list(stat_df.columns):
        #             stat_df[col] = stat_df[col].round(1)
        #             stat_df[col] = stat_df[col].map(str)
        #             stat_df = fix_col(stat_df, col)

        #         stat_df = stat_df.loc[many_order, :]

        #         show_st_table(stat_df.rename_axis(rename_col_dict[id_col]).reset_index(), st_col=stat_col)

        #     st.markdown("---")

        #     cont_df = pd.DataFrame()

        #     cont_order = list()

        #     for index in list(many_df.index.values):
                
        #         lig = many_df.at[index, pharm_lig_col]
        #         cont_dict = str_to_dict(many_df.at[index, bound_lig_cont_col], return_int=True)

        #         cont_lst = cont_dict[lig]

        #         label = many_df.at[index, id_col]
        #         label_name = f"{label} (N={count_dict[label]})"

        #         if label_name not in list(cont_df.columns):
        #             cont_df[label_name] = 0

        #         for cont in cont_lst:
        #             if cont in cont_order:
        #                 cont_df.at[cont, label_name] += 1
        #             else:
        #                 cont_df.at[cont, label_name] = 1
        #                 cont_order.append(cont)

        #     cont_df = cont_df.fillna(0)

        #     for col in list(cont_df.columns):
        #         cont_df[col] = (cont_df[col] / count_dict[col.split(" (")[0]]) * 100
        #         cont_df[col] = cont_df[col].round(1)
        #         cont_df[col] = cont_df[col].map(str)
        #         cont_df = fix_col(cont_df, col)
        #         cont_df[col] += "%"
        #         cont_df[col] = cont_df[col].replace("0%","")

        #     cont_df = cont_df.loc[sorted(cont_order), [f"{x} (N={count_dict[x]})" for x in many_order if x in list(count_dict.keys())]]

        #     cont_df = cont_df.rename_axis('Residue Contact').reset_index()

        #     left_table_col, right_table_col = st.columns(2)

        #     right_table_col.markdown("##### Site Comparison")

        #     show_st_table(cont_df, st_col=right_table_col)

        #     for table_col in [y32_name, y71_name, sw1_name, sw2_name]:

        #         left_table_col.markdown(f"##### {rename_col_dict[table_col]}")

        #         loop_df = (
        #             pd.pivot_table(
        #                 data=rename_st_cols(many_df),
        #                 index=[rename_col_dict[table_col]],
        #                 columns=rename_col_dict[id_col],
        #                 values=rename_col_dict[pdb_id_col],
        #                 aggfunc="nunique",
        #                 margins=True,
        #             )
        #             .fillna("")
        #         )

        #         for col in list(loop_df.columns):
        #             loop_df[col] = loop_df[col].map(str)
        #             loop_df = fix_col(loop_df, col)

        #         loop_df = reorder_st_cols(loop_df, table_col, id_col)

        #         loop_df = loop_df.reset_index()

        #         show_st_table(loop_df, st_col=left_table_col)

       
    with st.expander("Entries Table", expanded=False):

        st.markdown(f"##### Entries Table")

        show_df = rename_st_cols(mut_df)

        del show_df[rename_col_dict[pdb_id_col]]
        del show_df[rename_col_dict[modelid_col]]

        show_st_dataframe(show_df)

        entries_file_name = st.text_input(
            label="Entries File Name",
            value=f"{rascore_str}_{entry_table_file}",
        )
        download_st_df(show_df, entries_file_name, "Download Entries Table")            

    write_st_end()


from src.download.modules import *
from src.download.lookfilesinside import look_what_is_inside


def left_to_renumber_mmCIF(default_input_path_to_mmCIF=current_directory + "/mmCIF",
                           default_output_path_to_mmCIF=current_directory + "/output_mmCIF"):
    without_already_renumbered_mmCIF = list()
    # output_mmCIF_files_were_found_4Char = set()
    # input_mmCIF_files_were_found_4Char = set()
    output_mmCIF_files_were_found_set = set()
    input_mmCIF_files_were_found_set = set()

    mmCIF_files_were_found = look_what_is_inside("mmCIF", default_input_path_to_mmCIF=default_input_path_to_mmCIF)
    output_mmCIF_files_were_found = look_what_is_inside("output_mmCIF", default_output_path_to_mmCIF=default_output_path_to_mmCIF)

    for output_mmCIF_file in output_mmCIF_files_were_found:
        output_mmCIF_files_were_found_set.add(output_mmCIF_file)
    for input_mmCIF_file in mmCIF_files_were_found:
        input_mmCIF_files_were_found_set.add(input_mmCIF_file)

    set_difference = input_mmCIF_files_were_found_set - output_mmCIF_files_were_found_set

    for mmCIF_file in mmCIF_files_were_found:
        if mmCIF_file in set_difference:
            without_already_renumbered_mmCIF.append(mmCIF_file)

    return without_already_renumbered_mmCIF


def left_to_renumber_PDB(default_input_path_to_PDB=current_directory + "/PDB",
                         default_output_path_to_PDB=current_directory + "/output_PDB"):
    without_already_renumbered_PDB = list()
    output_PDB_files_were_found_4Char = set()
    input_PDB_files_were_found_4Char = set()

    input_PDB_files_were_found = look_what_is_inside("PDB", default_input_path_to_PDB=default_input_path_to_PDB)
    output_PDB_files_were_found = look_what_is_inside("output_PDB", default_output_path_to_PDB=default_output_path_to_PDB)

    for output_PDB_file in output_PDB_files_were_found:
        output_PDB_files_were_found_4Char.add(output_PDB_file[:4])
    for input_PDB_file in input_PDB_files_were_found:
        input_PDB_files_were_found_4Char.add(input_PDB_file[3:7])

    set_difference = input_PDB_files_were_found_4Char - output_PDB_files_were_found_4Char
    list_difference = list(set_difference)

    for PDB_id in list_difference:
        without_already_renumbered_PDB.append("pdb" + PDB_id + ".ent.gz")

    return without_already_renumbered_PDB


""" usage of left_to_renumber_mmCIF()"""
# left_to_renumber_mmCIF()
# left_to_renumber_PDB()



def log_writer(resulting):
    with open("log_corrected.tsv", "w") as f:
        comp_uni_human_uni_PDBid = list()
        pdb_id_set = set()
        formated_item = (
            format("SP", "<3")
            + format("PDB_id", "<7")
            + format("chain_PDB", "<12")
            + format("chain_auth", "<12")
            + format("UniProt", "<20")
            + format("SwissProt", "<20")
            + format("uni_len", ">10")
            + format("chain_len", ">10")
            + format("renum", ">10")
            + format("5k_or_50k", ">10")
        )
        f.write("%s\n" % formated_item)

        for n in resulting:
            if type(n) == list:
                for z in n:
                    if type(z) == list:
                        if z[0][-1] == "*":
                            try:
                                formated_item = (
                                    format("*", "<3")
                                    + format(z[0][:4], "<7")
                                    + format(z[1], "<12")
                                    + format(z[2], "<12")
                                    + format(z[3], "<20")
                                    + format(z[4], "<20")
                                    + format(z[5], ">10")
                                    + format(z[6], ">10")
                                    + format(z[7], ">10")
                                    + format(z[8], ">10")
                                )
                                pdb_id_set.add(z[0][:4])
                                comp_uni_human_uni_PDBid.append((z[3], z[4], z[0][:4]))
                            except Exception:
                                print(z)
                        else:
                            try:
                                formated_item = (
                                    format("+", "<3")
                                    + format(z[0], "<7")
                                    + format(z[1], "<12")
                                    + format(z[2], "<12")
                                    + format(z[3], "<20")
                                    + format(z[4], "<20")
                                    + format(z[5], ">10")
                                    + format(z[6], ">10")
                                    + format(z[7], ">10")
                                    + format(z[8], ">10")
                                )
                                pdb_id_set.add(z[0])
                                comp_uni_human_uni_PDBid.append((z[3], z[4], z[0][:4]))
                            except Exception:
                                print(z)
                        f.write("%s\n" % formated_item)

    uniq_comp_uni_human_uni_PDBid_translation = set()
    for n in comp_uni_human_uni_PDBid:
        if n[0] == "-":
            continue
        uniq_comp_uni_human_uni_PDBid_translation.add(n)

    with open("log_translator.tsv", "w") as file_handle:
        for listitem in uniq_comp_uni_human_uni_PDBid_translation:
            file_handle.write(
                listitem[0] + " " + listitem[1] + " " + listitem[2] + "\n"
            )



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
from tqdm import tqdm
import concurrent.futures

from ..functions.chem import get_lig_smiles, is_lig_match, get_lig_simi
from ..functions.lst import lst_unique, lst_to_str, res_to_lst, type_lst, sort_lst
from ..functions.lig import lig_col_lst, lig_lst_dict
from ..functions.coord import (
    load_coord,
    get_resname,
    get_neighbors,
    get_resnum,
    get_residues,
    get_reschainid,
    get_residue_cont,
    is_aa,
    is_het,
)
from ..functions.col import (
    core_path_col,
    modelid_col,
    chainid_col,
    pharm_lig_col,
    pharm_lig_site_col,
    pharm_lig_match_col,
    pharm_lig_smiles_col,
    bound_lig_cont_col
)
from ..functions.table import get_df_at_index, fix_val
from ..functions.path import save_table


def build_lig_df(
    df,
    index,
    lig_dir=None,
    site_dict=None,
    match_dict=None,
    other_site="Other",
    other_match="Unclassified",
    none_site="None",
    none_match="None",
    max_site_dist=4,
    coord_path_col=None,
):

    if coord_path_col is None:
        coord_path_col = core_path_col

    df = get_df_at_index(df, index)

    coord_path = df.at[index, coord_path_col]
    modelid = df.at[index, modelid_col]
    chainid = df.at[index, chainid_col]

    lig_dict = dict()

    for lig_col in lig_col_lst:

        lig_dict[lig_col] = list()

    structure = load_coord(coord_path)
    neighbors = get_neighbors(structure[fix_val(modelid, return_int=True)][chainid])

    cont_lst = list()
    site_lst = list()
    match_lst = list()
    smiles_lst = list()

    for residue in get_residues(structure):
        if is_het(residue):
            resid_cont_lst = lst_unique(
                [
                    get_resnum(x)
                    for x in get_residue_cont(
                        neighbors,
                        residue,
                        max_dist=max_site_dist,
                        level="R",
                    )
                    if (get_resnum(x) < 50000)
                    and (get_reschainid(x) == chainid)
                    and (is_aa(x))
                ]
            )

            if len(resid_cont_lst) > 0:
                resname = get_resname(residue)
                lig_class = False
                is_pharm = False
                for lig_col in lig_col_lst:
                    if resname in lig_lst_dict[lig_col]:
                        lig_dict[lig_col].append(resname)
                        lig_class = True
                        if lig_col == pharm_lig_col:
                            is_pharm = True

                if not lig_class:
                    lig_dict[pharm_lig_col].append(resname)
                    is_pharm = True


                if is_pharm:
                    resid_cont_str = resname
                    resid_cont_str += ':'
                    resid_cont_str += lst_to_str(resid_cont_lst)
                    cont_lst.append(resid_cont_str)
                    
                    resid_smiles_str = resname
                    resid_smiles_str += ':'
                    resid_smiles_str += get_lig_smiles(resname, lig_dir=lig_dir)
                    smiles_lst.append(resid_smiles_str)

                    if site_dict is not None:
                        max_cont = 0
                        pharm_site = other_site
                        for site_name, site_cont_lst in site_dict.items():
                            site_cont = len([x for x in site_cont_lst if x in resid_cont_lst])
                            if site_cont > max_cont:
                                pharm_site = site_name
                                max_cont = site_cont
                        site_lst.append(pharm_site)

                    if match_dict is not None:
                        if pharm_site in list(match_dict.keys()):
                            pharm_match = other_match
                            match_name_lst = list()
                            match_simi_lst = list()
                            for match_name, query_lst in match_dict[pharm_site].items():
                                is_match = is_lig_match(
                                    resname, query_lst, lig_dir=lig_dir
                                )
                                if is_match:
                                    for query in query_lst:
                                        match_name_lst.append(match_name)
                                        match_simi_lst.append(
                                            get_lig_simi(
                                                resname, query, lig_dir=lig_dir
                                            )
                                        )
                            if len(match_name_lst) > 0:
                                pharm_match = match_name_lst[
                                    match_simi_lst.index(max(match_simi_lst))
                                ]
                            pharm_match = f"{pharm_site}.{pharm_match}"
                        else:
                            pharm_match = other_site
                        match_lst.append(pharm_match)

    for lig_col in lig_col_lst:
        df.at[index, lig_col] = lst_to_str(sort_lst(lst_unique(lig_dict[lig_col])))

    df.at[index, pharm_lig_site_col] = lst_to_str(
        sort_lst(lst_unique(site_lst)), join_txt="|", empty=none_site
    )
    df.at[index, pharm_lig_match_col] = lst_to_str(
        sort_lst(lst_unique(match_lst)), join_txt="|", empty=none_match
    )
    df.at[index, pharm_lig_smiles_col] = lst_to_str(smiles_lst, join_txt="|")
    df.at[index, bound_lig_cont_col] = lst_to_str(cont_lst, join_txt="|")

    return df


def annot_lig(
    df,
    lig_table_path=None,
    lig_dir=None,
    site_dict=None,
    match_dict=None,
    other_site="Other",
    other_match="Unclassified",
    none_site="None",
    none_match="None",
    max_site_dist=4,
    coord_path_col=None,
    num_cpu=1,
):

    df = df.reset_index(drop=True)

    if site_dict is not None:
        for site_name, site_cont in site_dict.items():
            site_dict[site_name] = res_to_lst(site_cont)

    if match_dict is not None:
        for site_name in list(match_dict.keys()):
            for match_name, query_lst in match_dict[site_name].items():
                match_dict[site_name][match_name] = type_lst(query_lst)

    lig_df = pd.DataFrame()

    if num_cpu == 1:
        for index in tqdm(
            list(df.index.values), desc="Annotating ligands", position=0, leave=True
        ):

            lig_df = pd.concat(
                [
                    lig_df,
                    build_lig_df(
                        df,
                        index,
                        lig_dir=lig_dir,
                        site_dict=site_dict,
                        match_dict=match_dict,
                        other_site=other_site,
                        other_match=other_match,
                        none_site=none_site,
                        none_match=none_match,
                        max_site_dist=max_site_dist,
                        coord_path_col=coord_path_col,
                    ),
                ],
                sort=False,
            )
    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:
            job_lst = [
                executor.submit(
                    build_lig_df,
                    df,
                    index,
                    site_dict=site_dict,
                    match_dict=match_dict,
                    other_site=other_site,
                    other_match=other_match,
                    none_site=none_site,
                    none_match=none_match,
                    max_site_dist=max_site_dist,
                    coord_path_col=coord_path_col,
                )
                for index in list(df.index.values)
            ]

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Annotating ligands",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):

                lig_df = pd.concat([lig_df, job.result()], sort=False)

    lig_df = lig_df.reset_index(drop=True)

    print("Annotated ligands!")

    if lig_table_path is not None:
        save_table(lig_table_path, lig_df)
    else:
        return lig_df



from src.download.modules import *


def SIFTS_tree_parser(handle_SIFTS):
    tree = ET.parse(handle_SIFTS)
    root = tree.getroot()

    crossRefDb_list = list()
    PDBe_val_tuples_in_list = list()
    PDBe_val_tuples_in_list_for_Uni = list()
    PDBe_val_tuples_in_list_for_PDB = list()
    PDB_val_tuples_in_list = list()
    UniProt_val_tuple_in_list = list()
    UniProtdbAccessionId_list = list()
    UniProt_conversion_dict = dict()

    for entity in root:
        if entity.tag.endswith("entity"):
            entity_chainID_list = list(entity.attrib.items())
            if entity_chainID_list[0][0] == "type" and entity_chainID_list[0][1] == "protein":
                for segment in entity:
                    for listResidue in segment:
                        if listResidue.tag.endswith("listMapRegion"):
                            for mapRegion in listResidue:
                                for db in mapRegion:
                                    dbSource_UniProt = list(db.attrib.items())
                                    if "dbSource" == dbSource_UniProt[0][0] and "UniProt" == dbSource_UniProt[0][1]:
                                        if db.text is None:
                                            UniProt = dbSource_UniProt[2][1]
                                        else:
                                            Human_readable = db.text
                                            try:
                                                UniProt_conversion_dict[UniProt] = Human_readable
                                            except NameError:
                                                UniProt_conversion_dict["UniProt"] = "NameError"

                        for residue in listResidue:
                            key_val_tuples_in_list_parent = list(residue.attrib.items())
                            if key_val_tuples_in_list_parent[0][0] == "dbSource" and key_val_tuples_in_list_parent[0][1] == "PDBe":
                                PDBe_val_tuples_in_list.append((key_val_tuples_in_list_parent[2][1],
                                                                key_val_tuples_in_list_parent[3][1],
                                                                entity_chainID_list[1][1]))

                                for crossRefDb in residue:
                                    crossRefDb_list.append(crossRefDb.attrib)
                                    key_val_tuples_in_list_child = list(crossRefDb.attrib.items())

                                    if key_val_tuples_in_list_child[0][0] == "dbSource" and key_val_tuples_in_list_child[0][1] == "PDB":
                                        PDB_val_tuples_in_list.append((key_val_tuples_in_list_child[3][1],
                                                                       key_val_tuples_in_list_child[4][1],
                                                                       key_val_tuples_in_list_child[5][1]))
                                        PDBe_val_tuples_in_list_for_PDB.append((key_val_tuples_in_list_parent[2][1],
                                                                                key_val_tuples_in_list_parent[3][1],
                                                                                entity_chainID_list[1][1]))

                                    if key_val_tuples_in_list_child[0][0] == "dbSource" and key_val_tuples_in_list_child[0][1] == "UniProt":
                                        UniProt_val_tuple_in_list.append((key_val_tuples_in_list_child[3][1],
                                                                          key_val_tuples_in_list_child[4][1],
                                                                          entity_chainID_list[1][1]))
                                        PDBe_val_tuples_in_list_for_Uni.append((key_val_tuples_in_list_parent[2][1],
                                                                                key_val_tuples_in_list_parent[3][1],
                                                                                entity_chainID_list[1][1]))
                                        UniProtdbAccessionId_list.append(key_val_tuples_in_list_child[2][1])

    tuple_PDBe_for_PDB_and_tuple_PDB = list(zip(PDBe_val_tuples_in_list_for_PDB, PDB_val_tuples_in_list))
    tuple_PDBe_for_UniProt_and_tuple_UniProt = list(zip(PDBe_val_tuples_in_list_for_Uni, UniProt_val_tuple_in_list, UniProtdbAccessionId_list))

    return [tuple_PDBe_for_PDB_and_tuple_PDB, tuple_PDBe_for_UniProt_and_tuple_UniProt, UniProt_conversion_dict]



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import os
import shutil
import gzip
import json
import pandas as pd
import numpy as np

from .table import order_cols, order_rows, get_col_order

rascore_str = "rascore"
build_str = "build"
classify_str = "classify"
cluster_str = "cluster"
plot_str = "plot"

util_str = "util"
pipelines_str = "pipelines"
data_str = "data"
pages_str = "pages"
functions_str = "functions"

pdbaa_str = "pdbaa"
core_str = "core"
rcsb_str = "rcsb"
renum_str = "renum"
rcsb_assembly_str = "rcsb_assembly"
renum_assembly_str = "renum_assembly"
sifts_str = "sifts"
edia_str = "edia"
seq_str = "sequence"
interf_str = "interface"
pocket_str = "pocket"
lig_str = "ligand"
eds_str = "eds"
map_str = "2mFo-DFc"
diff_str = "mFo-DFc"
sup_str = "super"


def path_exists(path):

    exists = False
    if path is not None:
        if os.path.isfile(path):
            exists = True
        elif os.path.isdir(path):
            exists = True

    return exists


def append_path(path):

    if not path_exists(path):
        os.makedirs(path)


def get_dir_name(dir_path):

    if "/" in dir_path:
        dir_name = dir_path.rsplit("/", 1)[0]
    else:
        dir_name = os.getcwd()

    return dir_name


def get_file_name(path):

    if "/" in path:
        file_name = path.rsplit("/", 1)[1]
    else:
        file_name = path

    return file_name


def append_file_path(path):

    append_path(get_dir_name(path))


def delete_path(path):

    if path_exists(path):
        if os.path.isfile(path):
            os.remove(path)
        elif os.path.isdir(path):
            shutil.rmtree(path)


def copy_path(source_path, dest_path):

    if path_exists(dest_path):
        delete_path(dest_path)
    shutil.copyfile(source_path, dest_path)


def save_table(path, df, sep="\t", header=True, index=False, fillna="None"):

    append_file_path(path)

    df = df.fillna(fillna)

    df = order_cols(df, get_col_order(df))

    df = order_rows(df)

    df.to_csv(path, sep=sep, header=header, index=index)


def load_table(path, sep="\t"):

    if path_exists(path):
        df = pd.read_csv(path, sep=sep, dtype=str)
        df = order_rows(df)
    else:
        df = None

    return df


def save_matrix(path, matrix, delim=","):

    append_file_path(path)

    np.savetxt(path, matrix, delimiter=delim)


def load_matrix(path, delim=","):

    if path_exists(path):
        matrix = np.loadtxt(path, delimiter=delim)
    else:
        matrix = None

    return matrix


def save_lst(path, val_lst):

    append_file_path(path)

    with open(path, "w") as file:
        for val in val_lst:
            file.write(f"{val}\n")


def load_lst(path):

    if path_exists(path):
        with open(path, "r") as file:
            line_lst = file.read().splitlines()
    else:
        line_lst = None

    return line_lst


def save_json(path, json_dict):

    append_file_path(path)

    with open(path, "w") as file:
        json.dump(json_dict, file)


def load_json(path):

    if path_exists(path):
        with open(path, "r") as file:
            json_dict = json.load(file)
    else:
        json_dict = None

    return json_dict


def unzip_file(in_path, out_path=None):

    if out_path is None:
        out_path = in_path.replace(".gz", "")

    with gzip.open(in_path, "rb") as file_in:
        with open(out_path, "wb") as file_out:
            shutil.copyfileobj(file_in, file_out)


def search_dir(dir_path, file_str):

    return [x for x in os.listdir(dir_path) if file_str in x]


def get_dir_path(dir_str=None, dir_path=None):

    if dir_path is None:
        dir_path = os.getcwd()

    if dir_str is not None:
        dir_path += f"/{dir_str}"

    return dir_path


def get_file_path(file_name, dir_str=None, dir_path=None, pre_str=True):

    file_path = get_dir_path(dir_str=dir_str, dir_path=dir_path)
    file_path += "/"
    if pre_str and dir_str != None:
        file_path += dir_str
        file_path += "_"
    file_path += file_name

    return file_path


def modify_coord_path(path, return_pdb=False, add_h=False):

    if return_pdb:
        path = path.replace(".cif", ".pdb")
        if add_h:
            path = path.replace(".pdb", ".h.pdb")

    return path


def get_core_path(
    pdb_code, chainid, modelid=None, dir_path=None, return_pdb=False, add_h=False
):
    pdb_id = f"{pdb_code}{chainid}"

    if modelid is not None:
        pdb_id += str(modelid)

    return modify_coord_path(
        get_file_path(
            f"{pdb_id}_{core_str}.cif",
            dir_str=core_str,
            dir_path=dir_path,
            pre_str=False,
        ),
        return_pdb=return_pdb,
        add_h=add_h,
    )


def get_rcsb_path(pdb_code, dir_path=None):

    return get_file_path(
        f"{pdb_code}.cif.gz", dir_str=rcsb_str, dir_path=dir_path, pre_str=False
    )


def get_sifts_path(pdb_code, dir_path=None):

    return get_file_path(
        f"{pdb_code}.xml.gz", dir_str=sifts_str, dir_path=dir_path, pre_str=False
    )


def get_renum_path(pdb_code, dir_path=None):

    return get_file_path(
        f"{pdb_code}_{renum_str}.cif",
        dir_str=renum_str,
        dir_path=dir_path,
        pre_str=False,
    )


def get_seq_path(uniprot_acc, dir_path=None):

    return get_file_path(
        f"{uniprot_acc}.fasta", dir_str=seq_str, dir_path=dir_path, pre_str=False
    )


def get_edia_path(pdb_code, dir_path=None):

    return get_file_path(
        f"{pdb_code}_{edia_str}.csv", dir_str=edia_str, dir_path=dir_path, pre_str=False
    )


def get_lig_path(lig, dir_path=None):

    return get_file_path(
        f"{lig}_{lig_str}.sdf", dir_str=lig_str, dir_path=dir_path, pre_str=False
    )


def get_eds_map_path(pdb_code, dir_path=None):

    return get_file_path(
        f"{pdb_code}_{map_str}.ccp", dir_str=map_str, dir_path=dir_path, pre_str=False
    )


def get_eds_diff_path(pdb_code, dir_path=None):

    return get_file_path(
        f"{pdb_code}_{diff_str}.ccp4",
        dir_str=diff_str,
        dir_path=dir_path,
        pre_str=False,
    )


def get_interf_path(pdb_code, chainid, interf, dir_path=None, return_pdb=False):

    return modify_coord_path(
        get_file_path(
            f"{pdb_code}_{chainid}_{interf}_{interf_str}.cif",
            dir_str=interf_str,
            dir_path=dir_path,
            pre_str=False,
        ),
        return_pdb=return_pdb,
    )


def get_pocket_path(pdb_id, pocket, dir_path=None):

    return get_file_path(
        f"{pdb_id}_{pocket}_{pocket_str}.pdb",
        dir_str=pocket_str,
        dir_path=dir_path,
        pre_str=False,
    )


def get_neighbor_path(file_path, dir_str, neighbor_str):

    dir_path = get_dir_name(file_path)
    dir_path = dir_path.split(dir_str)[0]
    dir_path += neighbor_str

    return dir_path



def short_usage_messenger():
    return """
PDB.py
optional arguments:
-h, --help            show this help message and exit

-rftf text_file_with_PDB.tsv, --renumber_from_text_file text_file_with_PDB.tsv
This option will download and renumber specified files
usage $ python3 PDB.py -rftf text_file_with_PDB_in_it.tsv -mmCIF
usage $ python3 PDB.py -rftf text_file_with_PDB_in_it.tsv -PDB
usage $ python3 PDB.py -rftf text_file_with_PDB_in_it.tsv -mmCIF_assembly
usage $ python3 PDB.py -rftf text_file_with_PDB_in_it.tsv -PDB_assembly
usage $ python3 PDB.py -rftf text_file_with_PDB_in_it.tsv -all

-rfla [6dbp 3v03 2jit ...], --renumber_from_list_of_arguments [6dbp 3v03 2jit ...]
This option will download and renumber specified files
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -mmCIF
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -PDB
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -mmCIF_assembly
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -PDB_assembly
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -all

-redb, --renumber_entire_database
This option will download and renumber entire PDB database in PDB or/and mmCIF format
usage $ python3 PDB.py -redb -mmCIF
usage $ python3 PDB.py -redb -PDB
usage $ python3 PDB.py -redb -mmCIF_assembly
usage $ python3 PDB.py -redb -PDB_assembly
usage $ python3 PDB.py -redb -all

for more info --help

Roland Dunbrack's Lab
Fox Chase Cancer Center
Philadelphia, PA
2020
        """



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from rdkit.Chem import Draw

from rdkit import Chem
from rdkit.Chem import (
    rdFMCS,
    AllChem,
    PyMol,
)

from tqdm import tqdm

from .lst import type_lst
from .color import get_rgb, change_hex_alpha, get_lst_colors
from .download import download_file
from .path import path_exists, get_lig_path, delete_path,lig_str
from .url import lig_expo_url


def download_lig(lig, lig_dir=None, check=True):

    lig_path = get_lig_path(lig, dir_path=lig_dir)

    download_file(
        f"{lig_expo_url}{lig[0]}/{lig}/{lig}_model.sdf", lig_path, check=check
    )

    return lig_path


def load_lig(lig, lig_dir=None, tries=3):

    lig_path = get_lig_path(lig, dir_path=lig_dir)

    if not path_exists(lig_path):
        lig_path = download_lig(lig, lig_dir=lig_dir)

    try_load = True
    count = 0
    while try_load:
        try:
            mol = Chem.MolFromMolFile(lig_path)
            try_load = False
        except:
            lig_path = download_lig(lig, lig_dir=lig_dir, check=False)
            count += 1
            if tries == count:
                try_load = False

    return mol


def get_lig_smiles(lig, lig_dir=None,tries=3):

    mol = lig
    if type(lig) == str:
        smiles = mol
        if len(lig) <= 3:
            mol = load_lig(lig, lig_dir=lig_dir)
    if type(mol) != str:
        try_load = True
        count = 0
        while try_load:
            try:
                smiles = Chem.MolToSmiles(mol)
                try_load = False
            except:
                lig_path = get_lig_path(lig, dir_path=lig_dir)
                delete_path(lig_path)
                mol = load_lig(lig, lig_dir=lig_dir,tries=tries)
                count += 1
                if tries == count:
                    try_load = False

    return smiles


def get_lig_mol(lig, lig_dir=None, stereo=True, tries=3):

    mol = lig
    if type(lig) == str:
        smiles = mol
        if len(lig) <= 3:
            smiles = get_lig_smiles(lig, lig_dir=lig_dir)

        try_load = True
        count = 0
        while try_load: 
            try:  
                mol = Chem.MolFromSmiles(smiles)
                try_load = False
            except:
                lig_path = get_lig_path(lig, dir_path=lig_dir)
                delete_path(lig_path)
                smiles = get_lig_smiles(lig, lig_dir=lig_dir,tries=tries)
                count += 1
                if tries == count:
                    try_load = False

    if not stereo:
        AllChem.Compute2DCoords(mol)

    return mol


def get_lig_simi(lig_1, lig_2, lig_dir=None):

    mol_1 = Chem.RDKFingerprint(get_lig_mol(lig_1, lig_dir=lig_dir))
    mol_2 = Chem.RDKFingerprint(get_lig_mol(lig_2, lig_dir=lig_dir))

    return Chem.DataStructs.FingerprintSimilarity(mol_1, mol_2)


def is_lig_match(lig, matches, lig_dir=None, return_total=False):

    mol = get_lig_mol(lig, lig_dir=lig_dir)

    match_lst = type_lst(matches)

    total_match = [mol.HasSubstructMatch(get_lig_mol(x, lig_dir=lig_dir)) for x in match_lst]

    if return_total:
        return len([x for x in total_match if x == True])
    else:
        if True in total_match:
            return True
        else:
            return False


def get_lig_mcs(
    lig_lst,
    lig_dir=None,
    min_simi=0.9,
    compare_bonds=False,
    compare_exact=False,
    match_valence=False,
    match_rings=False,
    complete_rings=False,
):

    mol_lst = list()
    for lig in tqdm(
        lig_lst,
        desc="Loading ligands",
        position=0,
        leave=True,
    ):

        mol_lst.append(get_lig_mol(lig, lig_dir=lig_dir))

    if compare_bonds:
        if compare_exact:
            bond_compare = rdFMCS.BondCompare.CompareOrderExact
        else:
            bond_compare = rdFMCS.BondCompare.CompareOrder
    else:
        bond_compare = rdFMCS.BondCompare.CompareAny

    return Chem.MolFromSmarts(
        rdFMCS.FindMCS(
            mol_lst,
            threshold=min_simi,
            bondCompare=bond_compare,
            matchValences=match_valence,
            ringMatchesRingOnly=match_rings,
            completeRingsOnly=complete_rings,
        ).smartsString
    )


def get_lig_query_match(lig, query, lig_dir=None):

    mol = get_lig_mol(lig, lig_dir=lig_dir)
    query = get_lig_mol(query, lig_dir=lig_dir)

    atom_lst = list(mol.GetSubstructMatch(query))
    bond_lst = list()

    if len(atom_lst) > 0:

        bond_lst += [
            mol.GetBondBetweenAtoms(
                atom_lst[x.GetBeginAtomIdx()], atom_lst[x.GetEndAtomIdx()]
            ).GetIdx()
            for x in query.GetBonds()
        ]

    return atom_lst, bond_lst


def draw_lig_plot(
    ligs,
    lig_dir=None,
    lig_labels=None,
    highlight_querys=None,
    highlight_alpha=0.75,
    color_palette=None,
    n_cols=None,
    mol_pad=0.1,
    plot_height=3,
    plot_width=3,
    font_size=2,
):

    lig_lst = type_lst(ligs)
    highlight_query_lst = type_lst(highlight_querys)

    if type(lig_labels) == dict:
        label_lst = [lig_labels[x] for x in lig_lst]
    else:
        label_lst = type_lst(lig_labels)
    if len(label_lst) != len(lig_lst):
        label_lst = None

    mol_lst = [get_lig_mol(x, lig_dir=lig_dir) for x in lig_lst]

    final_atom_lst = list()
    final_bond_lst = list()
    final_atom_color_lst = list()
    final_bond_color_lst = list()

    if highlight_querys is not None:
        if type(color_palette) == dict:
            color_dict = color_palette
            for key, val in color_dict.items():
                color_dict[key] = get_rgb(change_hex_alpha(val, highlight_alpha))
        else:
            color_dict = get_lst_colors(
                highlight_query_lst,
                palette=color_palette,
                return_rgb=True,
                return_dict=True,
                alpha=highlight_alpha,
            )

    for mol in mol_lst:

        mol_atom_lst = list()
        mol_bond_lst = list()
        mol_atom_color_dict = dict()
        mol_bond_color_dict = dict()

        if highlight_querys is not None:
            for highlight_query in highlight_query_lst:

                atom_lst, bond_lst = get_lig_query_match(mol, highlight_query)

                if len(atom_lst) > 0:

                    mol_atom_lst += atom_lst
                    mol_bond_lst += bond_lst

                    for atom in atom_lst:
                        mol_atom_color_dict[atom] = color_dict[highlight_query]

                    for bond in bond_lst:
                        mol_bond_color_dict[bond] = color_dict[highlight_query]

        final_atom_lst.append(mol_atom_lst)
        final_bond_lst.append(mol_bond_lst)
        final_atom_color_lst.append(mol_atom_color_dict)
        final_bond_color_lst.append(mol_bond_color_dict)

    total_mols = len(mol_lst)
    if n_cols is None:
        n_cols = total_mols
    plot_size = (plot_width * 72, plot_height * 72)
    n_rows = total_mols // n_cols
    if total_mols % n_cols:
        n_rows += 1

    total_size = (n_cols * plot_size[0], n_rows * plot_size[1])
    draw = Draw.rdMolDraw2D.MolDraw2DSVG(
        total_size[0], total_size[1], plot_size[0], plot_size[1]
    )

    if n_cols is None:
        n_cols = len(mol_lst)

    draw.drawOptions().legendFontSize = font_size
    draw.drawOptions().padding = mol_pad
    draw.SetFontSize(1)
    draw.SetLineWidth(1)
    draw.DrawMolecules(
        mol_lst,
        legends=label_lst,
        highlightAtoms=final_atom_lst,
        highlightBonds=final_bond_lst,
        highlightAtomColors=final_atom_color_lst,
        highlightBondColors=final_bond_color_lst,
    )
    draw.FinishDrawing()
    return draw.GetDrawingText()


def show_pymol_lig(
    lig,
    lig_dir=None,
):

    lig_path = download_lig(lig, lig_dir=lig_dir)

    mv = PyMol.MolViewer()

    mv.LoadFile(lig_path, lig_str)

    mv.Zoom(lig_str)



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""


def get_pdbaa_pdb_code(desc):

    return desc[0:4].lower()


def get_pdbaa_chainid(desc):

    return desc[4:5]


def get_desc_item(desc, index):

    return str(desc).split(None, 8)[index]


def get_pdbaa_method(desc):

    return get_desc_item(desc, 3)


def get_pdbaa_resolution(desc):

    return get_desc_item(desc, 4)


def get_pdbaa_r_factor(desc):

    return get_desc_item(desc, 5)


def get_pdbaa_prot(desc):

    return str(get_desc_item(desc, 8)).split(" <")[0]


def get_pdbaa_swiss_id(desc):

    return str(desc.split("<")[1].split(">")[0].split("(")[0])


# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
from tqdm import tqdm

from ..functions.coord import load_coord, calc_atom_dist, wmhb_name, no_hb_name
from ..functions.lst import type_lst, lst_nums, lst_to_str
from ..functions.col import (
    get_dist_col,
    core_path_col,
    modelid_col,
    chainid_col,
    vect_1_col,
    vect_2_col,
    hb_status_col,
    hb_angle_1_col,
    hb_angle_2_col,
    wmhb_angle_col,
    outlier_col,
)
from ..functions.table import get_df_at_index
from ..functions.path import modify_coord_path, save_table


def get_index_dist(
    df,
    index,
    x_resids,
    y_resids,
    x_atomids=None,
    y_atomids=None,
    shared_resids=None,
    shared_atomids=None,
    atom_dist_col_lst=None,
    vect_1_col_lst=None,
    vect_2_col_lst=None,
    hb_status_col_lst=None,
    hb_angle_1_col_lst=None,
    hb_angle_2_col_lst=None,
    wmhb_angle_col_lst=None,
    outlier_col_lst=None,
    check_hb=False,
    use_h=False,
    hb_sc=True,
    hb_bb=True,
    min_hb_dist=2.0,
    max_hb_dist=3.2,
    min_wmhb_dist=2.0,
    max_wmhb_dist=3.0,
    min_hb_angle=90,
    max_hb_angle=180,
    min_wmhb_angle=80,
    max_wmhb_angle=140,
    coord_path_col=None,
):

    if coord_path_col is None:
        coord_path_col = core_path_col

    index_df = get_df_at_index(df, index)

    coord_path = index_df.at[index, coord_path_col]
    modelid = index_df.at[index, modelid_col]
    chainid = index_df.at[index, chainid_col]

    val_lst = [x_resids, y_resids]

    if x_atomids is not None:
        val_lst.append(x_atomids)
    if y_atomids is not None:
        val_lst.append(y_atomids)

    if shared_resids is not None:
        val_lst.append(shared_resids)
    if shared_atomids is not None:
        val_lst.append(shared_atomids)

    for i, val in enumerate(val_lst):
        val_lst[i] = type_lst(val)

    x_lst = lst_nums(0, len(val_lst[0]) - 1)

    if use_h:
        coord_path = modify_coord_path(coord_path, return_pdb=True, add_h=True)

    structure = load_coord(coord_path)

    for x in x_lst:

        resid_1 = val_lst[0][x]
        resid_2 = val_lst[1][x]

        if x_atomids is not None:
            atomid_1 = val_lst[2][x]
        else:
            atomid_1 = None

        if y_atomids is not None:
            atomid_2 = val_lst[3][x]
        else:
            atomid_2 = None

        if atom_dist_col_lst is None:
            cont_atom_dist_col = get_dist_col(
                resid_1,
                resid_2,
                x_atomid=lst_to_str(atomid_1, join_txt="+"),
                y_atomid=lst_to_str(atomid_2, join_txt="+"),
            )
        else:
            cont_atom_dist_col = atom_dist_col_lst[x]

        if vect_1_col_lst is None:
            cont_vect_1_col = get_dist_col(
                resid_1,
                resid_2,
                x_atomid=lst_to_str(atomid_1, join_txt="+"),
                y_atomid=lst_to_str(atomid_2, join_txt="+"),
                ext=vect_1_col,
            )
        else:
            cont_vect_1_col = vect_1_col_lst[x]

        if vect_2_col_lst is None:
            cont_vect_2_col = get_dist_col(
                resid_1,
                resid_2,
                x_atomid=lst_to_str(atomid_1, join_txt="+"),
                y_atomid=lst_to_str(atomid_2, join_txt="+"),
                ext=vect_2_col,
            )
        else:
            cont_vect_2_col = vect_2_col_lst[x]

        if hb_status_col_lst is None:
            cont_hb_status_col = get_dist_col(
                resid_1,
                resid_2,
                x_atomid=lst_to_str(atomid_1, join_txt="+"),
                y_atomid=lst_to_str(atomid_2, join_txt="+"),
                ext=hb_status_col,
            )
        else:
            cont_hb_status_col = hb_status_col_lst[x]

        if hb_angle_1_col_lst is None:
            cont_hb_angle_1_col = get_dist_col(
                resid_1,
                resid_2,
                x_atomid=lst_to_str(atomid_1, join_txt="+"),
                y_atomid=lst_to_str(atomid_2, join_txt="+"),
                ext=hb_angle_1_col,
            )
        else:
            cont_hb_angle_1_col = hb_angle_1_col_lst[x]

        if hb_angle_2_col_lst is None:
            cont_hb_angle_2_col = get_dist_col(
                resid_1,
                resid_2,
                x_atomid=lst_to_str(atomid_1, join_txt="+"),
                y_atomid=lst_to_str(atomid_2, join_txt="+"),
                ext=hb_angle_2_col,
            )
        else:
            cont_hb_angle_2_col = hb_angle_2_col_lst[x]

        if wmhb_angle_col_lst is None:
            cont_wmhb_angle_col = get_dist_col(
                resid_1,
                resid_2,
                x_atomid=lst_to_str(atomid_1, join_txt="+"),
                y_atomid=lst_to_str(atomid_2, join_txt="+"),
                ext=wmhb_angle_col,
            )
        else:
            cont_wmhb_angle_col = wmhb_angle_col_lst[x]

        if outlier_col_lst is None:
            cont_outlier_col = get_dist_col(
                resid_1,
                resid_2,
                x_atomid=lst_to_str(atomid_1, join_txt="+"),
                y_atomid=lst_to_str(atomid_2, join_txt="+"),
                ext=outlier_col,
            )
        else:
            cont_outlier_col = outlier_col_lst[x]

        resid_lst = [resid_1, resid_2]

        for i, resid in enumerate(resid_lst):
            if type(resid) == str:
                resid = df.at[index, resid]
            resid_lst[i] = resid

        result = calc_atom_dist(
            structure=structure,
            chainid_1=chainid,
            modelid_1=modelid,
            resid_1=resid_lst[0],
            atomid_1=atomid_1,
            chainid_2=chainid,
            modelid_2=modelid,
            resid_2=resid_lst[1],
            atomid_2=atomid_2,
            check_hb=check_hb,
            use_h=use_h,
            hb_sc=hb_sc,
            hb_bb=hb_bb,
            min_hb_dist=min_hb_dist,
            max_hb_dist=max_hb_dist,
            min_wmhb_dist=min_wmhb_dist,
            max_wmhb_dist=max_wmhb_dist,
            min_hb_angle=min_hb_angle,
            max_hb_angle=max_hb_angle,
            min_wmhb_angle=min_wmhb_angle,
            max_wmhb_angle=max_wmhb_angle,
            return_vect=True,
        )

        atom_dist = result[0]
        vect_1 = result[1]
        vect_2 = result[2]

        if check_hb:
            hb_status = result[3]
            hb_angle_1 = result[4]
            hb_angle_2 = result[5]
            wmhb_angle = result[6]
            outlier_status = result[7]

            if hb_status == wmhb_name:
                if shared_resids is not None:
                    shared_resid = val_lst[4][x]

                    if shared_atomids is not None:
                        shared_atomid = val_lst[5][x]
                    else:
                        shared_atomid = None

                    shared_result_1 = calc_atom_dist(
                        structure=structure,
                        chainid_1=chainid,
                        modelid_1=modelid,
                        resid_1=resid_lst[0],
                        atomid_1=atomid_1,
                        chainid_2=chainid,
                        modelid_2=modelid,
                        resid_2=shared_resid,
                        atomid_2=shared_atomid,
                        check_hb=check_hb,
                        use_h=use_h,
                        hb_sc=hb_sc,
                        hb_bb=hb_bb,
                        min_hb_dist=min_hb_dist,
                        max_hb_dist=max_hb_dist,
                        min_wmhb_dist=min_wmhb_dist,
                        max_wmhb_dist=max_wmhb_dist,
                        min_hb_angle=min_hb_angle,
                        max_hb_angle=max_hb_angle,
                        min_wmhb_angle=min_wmhb_angle,
                        max_wmhb_angle=max_wmhb_angle,
                        return_vect=True,
                    )
                    shared_result_2 = calc_atom_dist(
                        structure=structure,
                        chainid_1=chainid,
                        modelid_1=modelid,
                        resid_1=shared_resid,
                        atomid_1=shared_atomid,
                        chainid_2=chainid,
                        modelid_2=modelid,
                        resid_2=resid_lst[1],
                        atomid_2=atomid_2,
                        check_hb=check_hb,
                        use_h=use_h,
                        hb_sc=hb_sc,
                        hb_bb=hb_bb,
                        min_hb_dist=min_hb_dist,
                        max_hb_dist=max_hb_dist,
                        min_wmhb_dist=min_wmhb_dist,
                        max_wmhb_dist=max_wmhb_dist,
                        min_hb_angle=min_hb_angle,
                        max_hb_angle=max_hb_angle,
                        min_wmhb_angle=min_wmhb_angle,
                        max_wmhb_angle=max_wmhb_angle,
                        return_vect=True,
                    )

                    if (shared_result_1[3] == no_hb_name) or (
                        shared_result_2[3] == no_hb_name
                    ):
                        hb_status = no_hb_name
                        outlier_status = True

        index_df.at[
            index,
            cont_atom_dist_col,
        ] = atom_dist

        index_df.at[
            index,
            cont_vect_1_col,
        ] = vect_1

        index_df.at[
            index,
            cont_vect_2_col,
        ] = vect_2

        if check_hb:
            index_df.at[
                index,
                cont_hb_status_col,
            ] = hb_status

            index_df.at[
                index,
                cont_hb_angle_1_col,
            ] = hb_angle_1

            index_df.at[
                index,
                cont_hb_angle_2_col,
            ] = hb_angle_2

            index_df.at[
                index,
                cont_wmhb_angle_col,
            ] = wmhb_angle
            index_df.at[
                index,
                cont_outlier_col,
            ] = outlier_status

    return index_df


def build_dist_table(
    df,
    x_resids,
    y_resids,
    x_atomids=None,
    y_atomids=None,
    shared_resids=None,
    shared_atomids=None,
    dist_table_path=None,
    atom_dist_col_lst=None,
    hb_status_col_lst=None,
    hb_angle_1_col_lst=None,
    hb_angle_2_col_lst=None,
    wmhb_angle_col_lst=None,
    outlier_col_lst=None,
    check_hb=False,
    use_h=False,
    hb_sc=True,
    hb_bb=True,
    min_hb_dist=2.0,
    max_hb_dist=3.2,
    min_wmhb_dist=2.0,
    max_wmhb_dist=3.0,
    min_hb_angle=90,
    max_hb_angle=180,
    min_wmhb_angle=80,
    max_wmhb_angle=140,
    coord_path_col=None,
):

    df = df.reset_index(drop=True)

    dist_df = pd.DataFrame()

    for index in tqdm(
        list(df.index.values), desc="Building distance table", position=0, leave=True
    ):
        dist_df = pd.concat(
            [
                dist_df,
                get_index_dist(
                    df,
                    index,
                    x_resids,
                    y_resids,
                    x_atomids=x_atomids,
                    y_atomids=y_atomids,
                    shared_resids=shared_resids,
                    shared_atomids=shared_atomids,
                    atom_dist_col_lst=atom_dist_col_lst,
                    hb_status_col_lst=hb_status_col_lst,
                    hb_angle_1_col_lst=hb_angle_1_col_lst,
                    hb_angle_2_col_lst=hb_angle_2_col_lst,
                    wmhb_angle_col_lst=wmhb_angle_col_lst,
                    outlier_col_lst=outlier_col_lst,
                    check_hb=check_hb,
                    use_h=use_h,
                    hb_sc=hb_sc,
                    hb_bb=hb_bb,
                    min_hb_dist=min_hb_dist,
                    max_hb_dist=max_hb_dist,
                    min_wmhb_dist=min_wmhb_dist,
                    max_wmhb_dist=max_wmhb_dist,
                    min_hb_angle=min_hb_angle,
                    max_hb_angle=max_hb_angle,
                    min_wmhb_angle=min_wmhb_angle,
                    max_wmhb_angle=max_wmhb_angle,
                    coord_path_col=coord_path_col,
                ),
            ],
            sort=False,
        )

    dist_df = dist_df.reset_index(drop=True)

    print("Built distance table!")

    if dist_table_path is not None:
        save_table(dist_table_path, dist_df)
    else:
        return dist_df



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
import numpy as np
from tqdm import tqdm
import concurrent.futures

from ..functions.lst import type_lst
from ..functions.coord import load_cif_dict, search_cif_dict
from ..functions.path import save_table
from ..functions.table import merge_tables
from ..functions.col import (
    len_a_col,
    len_b_col,
    len_c_col,
    ang_a_col,
    ang_b_col,
    ang_g_col,
    rcsb_path_col,
    space_col,
    cf_col,
)

len_col_lst = [len_a_col, len_b_col, len_c_col]
ang_col_lst = [ang_a_col, ang_b_col, ang_g_col]
len_letter_lst = ["a", "b", "c"]
ang_letter_lst = ["alpha", "beta", "gamma"]


def get_path_len_ang_df(coord_path):

    df = pd.DataFrame()

    cif_dict = load_cif_dict(coord_path)

    df.at[0, rcsb_path_col] = coord_path

    if search_cif_dict(cif_dict, "_exptl.method") == "X-RAY DIFFRACTION":

        df.at[0, space_col] = search_cif_dict(
            cif_dict, "_symmetry.space_group_name_H-M"
        )

        for col, letter in zip(len_col_lst, len_letter_lst):

            df.at[0, col] = search_cif_dict(cif_dict, f"_cell.length_{letter}")

        for col, letter in zip(ang_col_lst, ang_letter_lst):

            df.at[0, col] = search_cif_dict(cif_dict, f"_cell.angle_{letter}")

    else:
        df.at[0, space_col] = "None"
        for col, letter in zip(len_col_lst, len_letter_lst):
            df.at[0, col] = 999.00
        for col, letter in zip(ang_col_lst, ang_letter_lst):
            df.at[0, col] = 999.00

    return df


def add_cf(df, min_simi=0.05):

    val_dict = dict()
    prev_dict = dict()

    col_lst = len_col_lst + ang_col_lst

    for index in tqdm(
        list(df.index.values), desc="Annotating crystal forms", position=0, leave=True
    ):

        curr_path = df.at[index, rcsb_path_col]
        curr_space = df.at[index, space_col]

        if curr_space == "None":
            df.at[index, cf_col] = "None"
        else:
            val_dict[curr_path] = {
                len_a_col: df.at[index, len_a_col],
                len_b_col: df.at[index, len_b_col],
                len_c_col: df.at[index, len_c_col],
                ang_a_col: df.at[index, ang_a_col],
                ang_b_col: df.at[index, ang_b_col],
                ang_g_col: df.at[index, ang_g_col],
            }

            if curr_space not in list(prev_dict.keys()):
                id = 1
                prev_dict[curr_space] = {id: type_lst(curr_path)}

            else:

                new_form = True

                for id in list(prev_dict[curr_space].keys()):

                    prev_path_lst = prev_dict[curr_space][id]

                    diff_form = False

                    for col in col_lst:

                        diff_lst = list()

                        for prev_path in prev_path_lst:

                            curr_val = float(val_dict[curr_path][col])
                            prev_val = float(val_dict[prev_path][col])

                            diff_lst.append(abs(prev_val - curr_val) / prev_val)

                        mean_diff = np.mean(diff_lst)

                        if mean_diff > min_simi:
                            diff_form = True
                            break

                    if not diff_form:

                        prev_dict[curr_space][id].append(curr_path)

                        new_form = False

                        break

                if new_form:

                    id = max(list(prev_dict[curr_space].keys())) + 1

                    prev_dict[curr_space][id] = type_lst(curr_path)

            df.at[index, cf_col] = f"{curr_space} (CF{id})"

    return df


def annot_cf(coord_paths, cf_table_path=None, min_simi=0.05, data=None, num_cpu=1):

    coord_path_lst = type_lst(coord_paths)

    df = pd.DataFrame()

    if num_cpu == 1:
        for coord_path in tqdm(
            coord_path_lst, desc="Getting crystal information", position=0, leave=True
        ):

            df = pd.concat([df, get_path_len_ang_df(coord_path)], sort=False)
    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:
            job_lst = [
                executor.submit(get_path_len_ang_df, coord_path)
                for coord_path in coord_path_lst
            ]

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Getting crystal information",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):

                df = pd.concat([df, job.result()], sort=False)

    df = df.reset_index(drop=True)

    df = add_cf(df, min_simi=min_simi)

    if data is not None:
        df_col_lst = list(data.columns)

        for col in [cf_col]:
            if col in df_col_lst:
                del data[col]

        df = merge_tables(df, data)

    print("Annotated crystal forms!")

    if cf_table_path is not None:
        save_table(cf_table_path, df)
    else:
        return df


# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
import numpy as np
import math
import itertools

from .lst import move_end_lst
from .col import (
    cluster_col,
    pdb_code_col,
    cf_col,
    nn_dist_col,
    constr_dist_col,
    entropy_col,
    common_col,
    occup_col,
    rama_col,
    rotamer_col,
    bb_seq_col,
    sc_seq_col,
    iso_col,
    dist_col_lst,
    interf_area_col,
    pocket_volume_col,
    pocket_score_col,
    mean_col,
    max_col,
)
from .table import (
    build_count_table,
    merge_tables,
    mask_equal,
    build_count_dict,
    build_col_count_dict,
    lst_col,
    lst_by_freq,
    get_val_index_lst,
    mask_matrix,
    get_col_most_common,
)


def is_noise(val):

    noise_status = False
    if type(val) == str:
        if 'Noise' in val:
            noise_status = True
    elif val == -1:
        noise_status = True

    return noise_status


def label_cluster(val, cluster_order):

    if is_noise(val):
        return "Noise"
    else:
        return str(cluster_order.index(val) + 1)


def order_clusters(df, order_col=None):

    temp_df = df.copy(deep=True)
    if order_col is not None:
        temp_df = temp_df.loc[:, [cluster_col, order_col]]
        temp_df = temp_df.drop_duplicates()

    cluster_lst = lst_col(temp_df, cluster_col)

    cluster_order = lst_by_freq(cluster_lst)

    for cluster in cluster_order:
        if is_noise(cluster):
            cluster_order = move_end_lst(cluster_order, "Noise")

    df[cluster_col] = df.apply(
        lambda x: label_cluster(x[cluster_col], cluster_order),
        axis=1,
    )

    return df


def renumber_clusters(df, min_samples, min_pdb=2, min_cf=1, order_col=None):

    count_dict = build_count_dict(lst_col(df, cluster_col, return_str=True))

    check_pdb = False
    if pdb_code_col in list(df.columns):
        pdb_dict = build_col_count_dict(
            df, cluster_col, col_lst=[pdb_code_col], return_str=True
        )
        check_pdb = True

    check_cf = False
    if cf_col in list(df.columns):
        cf_dict = build_col_count_dict(
            df, cluster_col, col_lst=[cf_col], return_str=True
        )
        check_cf = True

    df[cluster_col] = df[cluster_col].map(str)

    label_dict = dict()
    for cluster in lst_col(df, cluster_col, unique=True, return_str=True):

        label = cluster

        if count_dict[cluster] < min_samples:
            label = "Noise"

        if check_pdb:
            if pdb_dict[cluster] < min_pdb:
                label = "Noise"

        if check_cf:
            if cf_dict[cluster] < min_cf:
                label = "Noise"

        label_dict[cluster] = label

    df[cluster_col] = df[cluster_col].map(label_dict)

    df = order_clusters(df, order_col=order_col)

    return df


def get_index_dist_lst(matrix, index_lst, index=None):

    if index is not None:
        index_lst.remove(index)

    return matrix[index_lst, index]


def calc_dist_stat(dist_lst, method="mean", dec=2):

    if len(dist_lst) < 2:
        dist = dist_lst[0]
    else:
        if method == "mean":
            dist = np.mean(dist_lst)
        elif method == "max":
            dist = np.max(dist_lst)
        elif method == "min":
            dist = np.min(dist_lst)
        elif method == "median":
            dist = np.median(dist_lst)
        elif method == "std":
            dist = np.std(dist_lst)

    return round(dist, dec)


def calc_cutoff(dist_lst, dec=2):

    return round(
        calc_dist_stat(dist_lst, method="mean")
        + (2 * calc_dist_stat(dist_lst, method="std")),
        dec,
    )


def build_cutoff_dict(df, matrix, method="mean"):

    cluster_lst = lst_col(df, cluster_col, unique=True, return_str=True)

    cutoff_dict = dict()

    for cluster in cluster_lst:

        index_lst = get_val_index_lst(df, cluster_col, cluster)

        dist_lst = list()

        if len(index_lst) == 1:
            dist_lst.append(0.0)
        else:
            for index in index_lst:
                dist_lst.append(
                    calc_dist_stat(
                        get_index_dist_lst(matrix, index_lst, index=index),
                        method=method,
                    )
                )

        cutoff_dict[cluster] = calc_cutoff(dist_lst)

    return cutoff_dict


def apply_dist_constr(
    df,
    dist_col,
    matrix,
    min_samples=None,
    max_dist=None,
    constr_method=None,
    order_col=None,
):
    df[cluster_col] = df[cluster_col].map(str)

    df_index_lst = list(df.index.values)

    if max_dist is None:
        cutoff_dict = build_cutoff_dict(df, matrix, method=constr_method)

    pre_count_dict = build_col_count_dict(df, cluster_col)

    calc_dist = True
    while calc_dist:

        calc_dist = False

        for index in df_index_lst:

            cluster = df.at[index, cluster_col]

            index_lst = get_val_index_lst(df, cluster_col, cluster)

            if min_samples is not None:
                if not is_noise(cluster):
                    if len(index_lst) < min_samples:
                        cluster = "Noise"
                        df.at[index, cluster_col] = cluster
                        calc_dist = True

            if len(index_lst) == 1:
                dist = 0.0
            else:
                dist = calc_dist_stat(
                    get_index_dist_lst(matrix, index_lst, index=index),
                    method=constr_method,
                )

            df.at[index, dist_col] = dist

            if not is_noise(cluster):

                prune_cluster = False

                dist_cutoff = max_dist
                if max_dist is None:
                    dist_cutoff = cutoff_dict[cluster]

                if dist > dist_cutoff:
                    prune_cluster = True

                if prune_cluster:
                    df.at[index, cluster_col] = "Noise"
                    calc_dist = True

    post_count_dict = build_col_count_dict(df, cluster_col)

    pruned = 0
    for cluster in list(post_count_dict.keys()):
        if cluster != "Noise":
            pruned += pre_count_dict[cluster] - post_count_dict[cluster]

    df = order_clusters(df, order_col=order_col)

    return df, pruned


def prune_cluster_members(
    df,
    min_samples,
    matrix,
    max_nn_dist=None,
    constr_matrix=None,
    max_constr_dist=None,
    constr_method="mean",
    min_pdb=3,
    min_cf=1,
    order_col=None,
):

    df = renumber_clusters(
        df, min_samples, min_pdb=min_pdb, min_cf=min_cf, order_col=order_col
    )

    df, nn_pruned = apply_dist_constr(
        df,
        nn_dist_col,
        matrix,
        min_samples=min_samples,
        max_dist=max_nn_dist,
        constr_method="min",
        order_col=order_col,
    )

    if constr_matrix is not None:
        if max_constr_dist is None:
            max_constr_dist = build_cutoff_dict(df, constr_matrix, method=constr_method)

        df, constr_pruned = apply_dist_constr(
            df,
            constr_dist_col,
            constr_matrix,
            min_samples=min_samples,
            max_dist=max_constr_dist,
            constr_method=constr_method,
            order_col=order_col,
        )

    df = renumber_clusters(
        df,
        min_samples,
        min_pdb=min_pdb,
        min_cf=min_cf,
        order_col=order_col,
    )

    if constr_matrix is not None:
        return (df, nn_pruned, constr_pruned)
    else:
        return (df, nn_pruned)


def merge_clusters(df, matrix, max_merge_dist, merge_method="mean"):

    cluster_lst = [x for x in lst_col(df, cluster_col, unique=True) if not is_noise(x)]

    cluster_pairs = itertools.combinations(cluster_lst, 2)

    merge_dict = dict()

    for cluster_1, cluster_2 in cluster_pairs:

        index_1_lst = get_val_index_lst(df, cluster_col, cluster_1)
        index_2_lst = get_val_index_lst(df, cluster_col, cluster_2)

        if (
            calc_dist_stat(
                mask_matrix(matrix, index_1_lst, index_2_lst), method=merge_method
            )
            <= max_merge_dist
        ):
            merge_cluster = cluster_lst[
                min(cluster_lst.index(cluster_1), cluster_lst.index(cluster_2))
            ]
            merge_dict[cluster_1] = merge_cluster
            merge_dict[cluster_2] = merge_cluster

    df[cluster_col] = df[cluster_col].replace(merge_dict)

    return df


def calc_entropy(vals, dec=2):

    return round(
        (
            float(-1)
            * float(np.max(vals))
            / float(np.sum(vals))
            * (math.log10(float(np.max(vals)) / float(np.sum(vals))))
        ),
        dec,
    )


def calc_occupancy(vals, dec=2):

    return round(float(np.max(vals)) / float(np.sum(vals)), dec)


def add_dih_stats(temp_df, sum_df, col, index):

    count_dict = build_col_count_dict(temp_df, col)

    common = [x for x in get_col_most_common(temp_df, col) if "-" not in x]
    if len(common) == 0:
        common = "-"
    else:
        common = common[0]
    entropy = calc_entropy(list(count_dict.values()))
    occupancy = calc_occupancy(list(count_dict.values()))

    sum_df.at[index, f"{common_col}_{col}"] = common
    sum_df.at[index, f"{entropy_col}_{col}"] = entropy
    sum_df.at[index, f"{occup_col}_{col}"] = occupancy

    return sum_df


def dist_to_dih(dist, dec=2):

    return round(math.degrees(math.acos(1 - (float(dist) / 2))), dec)


def build_sum_table(df):

    df_col_lst = list(df.columns)

    dih_dist = False
    if rama_col in df_col_lst or rotamer_col in df_col_lst:
        dih_dist = True

    add_col_lst = [rama_col, rotamer_col, bb_seq_col, sc_seq_col, iso_col]

    df[cluster_col] = df[cluster_col].map(str)

    cluster_lst = lst_col(df, cluster_col, unique=True, return_str=True)

    sum_df = pd.DataFrame()

    stat_col_lst = (
        dist_col_lst + [interf_area_col] + [pocket_volume_col] + [pocket_score_col]
    )

    for cluster in cluster_lst:

        temp_df = mask_equal(df, cluster_col, cluster)

        index = cluster_lst.index(cluster)

        sum_df.at[index, cluster_col] = cluster

        if "Noise" not in cluster and "Outlier" not in cluster and "Disordered" not in cluster:
            for col in stat_col_lst:
                if col in df_col_lst:
                    dist_lst = lst_col(temp_df, col, return_float=True)

                    mean_dist = calc_dist_stat(dist_lst, method="mean")
                    max_dist = calc_dist_stat(dist_lst, method="max")

                    val_lst = [mean_dist, max_dist]

                    if dih_dist:
                        if col == nn_dist_col:
                            val_lst = [dist_to_dih(x) for x in val_lst]

                    sum_df.at[index, f"{col}_{mean_col}"] = val_lst[0]
                    sum_df.at[index, f"{col}_{max_col}"] = val_lst[1]

            for add_col in add_col_lst:
                if add_col in df_col_lst:
                    sum_df = add_dih_stats(temp_df, sum_df, add_col, index)

    sum_df = merge_tables(sum_df, build_count_table(df, cluster_col))

    sum_df = sum_df.fillna("-")

    return sum_df


# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)

import streamlit as st
from PIL import Image

from util.functions.path import get_file_path, get_dir_name, util_str, data_str

from util.pages.home_page import home_page
from util.pages.overview_page import overview_page
from util.pages.pdb_page import pdb_page
from util.pages.conformation_page import conformation_page
from util.pages.mutation_page import mutation_page
from util.pages.inhibitor_page import inhibitor_page
from util.pages.query_page import query_page
from util.pages.classify_page import classify_page


class MultiApp:
    def __init__(self):
        self.apps = []

    def add_app(self, title, func):
        self.apps.append({"title": title, "function": func})

    def run(self):
        img = Image.open(
            get_file_path(
                "rascore_logo.png",
                dir_path=f"{get_dir_name(__file__)}/{util_str}/{data_str}",
            ),
        )

        st.set_page_config(page_title="rascore", page_icon=img, layout="wide")

        st.sidebar.markdown("## Main Menu")
        app = st.sidebar.selectbox(
            "Select a Page", self.apps, format_func=lambda app: app["title"]
        )
        st.sidebar.markdown("---")
        app["function"]()


app = MultiApp()

app.add_app("Home Page", home_page)
app.add_app("Database Overview", overview_page)
app.add_app("Search PDB", pdb_page)
app.add_app("Explore Conformations", conformation_page)
app.add_app("Analyze Mutations", mutation_page)
app.add_app("Compare Inhibitors", inhibitor_page)
app.add_app("Query Database", query_page)
app.add_app("Classify Structures", classify_page)

app.run()
